"""
Reduced Hessian Generation

This module creates the reduced Hessian for use in various KIPET modules
"""
# Standard library imports
import os
from pathlib import Path

# Third party imports
import numpy as np
import pandas as pd
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.environ import Constraint, Param, Set, SolverFactory, Suffix
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix, triu
from scipy.sparse.linalg import spsolve

# KIPET library imports
from kipet.estimability_tools.parameter_handling import set_scaled_parameter_bounds

DEBUG = True


class ReducedHessian:
    """Class for handling the reduced hessian calculations in KIPET/NSD"""

    def __init__(self,
                 model_object,
                 kkt_method='k_aug',
                 global_param_name='d',
                 param_con_method='global',
                 scaled=False,
                 rho=10,
                 set_param_bounds=False,
                 parameter_set=None,
                 variable_name='P',
                 param_set_name='parameter_names',
                 current_set='current_p_set',
                 set_up_constraints=True,
                 use_duals=True,
                 global_constraint_name='fix_params_to_global',
                 file_number=None,
                 tmp_file='ipopt_output',
                 ):

        self.model_object = model_object
        self.kkt_method = kkt_method

        self.global_param_name = global_param_name
        self.global_constraint_name = global_constraint_name
        self.param_con_method = param_con_method
        self.scaled = scaled
        self.rho = rho
        self.set_param_bounds = set_param_bounds
        self.parameter_set = parameter_set
        self.variable_name = variable_name
        self.param_set_name = param_set_name
        self.current_set = current_set
        self.set_up_constraints = set_up_constraints
        self.use_duals = use_duals
        self.file_number = file_number
        self.tmp_file = tmp_file

        self.verbose = DEBUG
        self._parameter_set()
        
    def _parameter_set(self):
        
        if self.parameter_set is None:
            self.parameter_set = list(getattr(self.model_object, self.variable_name).keys())

    def get_tmp_file(self):
        """Read the output file generated by IPOPT

        :return: The tmp file with the correct number (on file)

        """
        file_tag = ''
        if self.file_number is not None:
            file_tag += f'_{self.file_number}'

        return self.tmp_file + file_tag

    def get_file_info(self):
        """This sets up the file dictionary (col, row, nl, and sol)

        :return: None

        """
        tmpfile_i = self.get_tmp_file()

        with open(tmpfile_i, 'r') as f:
            output_string = f.read()

        stub = output_string.split('\n')[0].split(',')[1][2:-4]

        nl_file = Path(stub + '.nl')
        col_file = Path(stub + '.col')
        row_file = Path(stub + '.row')
        sol_file = Path(stub + '.sol')

        self.sol_files = dict(
            nl=nl_file,
            col=col_file,
            row=row_file,
            sol=sol_file,
        )

        return None

    def delete_sol_files(self):
        """Removes the sol files, if found

        :return: None

        """
        if hasattr(self, 'sol_files'):

            for key, file in self.sol_files.items():
                file.unlink()

            del self.sol_files

        return None

    def get_kkt_info(self):
        """Takes the model and uses PyNumero or k_aug to get the jacobian and Hessian
        information as dataframes. This is done in place and does not return anything.

        kkt_data (dict): dictionary with the following structure:

                {
                'J': J,   # Jacobian
                'H': H,   # Hessian
                'var_ind': var_index_names, # Variable index
                'con_ind': con_index_names, # Constraint index
                'duals': duals, # Duals
                }

        :return: None

        """
        self.get_file_info()

        if self.kkt_method == 'pynumero':

            nlp = PyomoNLP(self.model_object)
            varList = nlp.get_pyomo_variables()
            conList = nlp.get_pyomo_constraints()
            duals = nlp.get_duals()

            J = nlp.extract_submatrix_jacobian(pyomo_variables=varList, pyomo_constraints=conList)
            H = nlp.extract_submatrix_hessian_lag(pyomo_variables_rows=varList, pyomo_variables_cols=varList)
            J = csc_matrix(J)

            var_index_names = [v.name for v in varList]
            con_index_names = [v.name for v in conList]

        elif self.kkt_method == 'k_aug':

            kaug = SolverFactory('k_aug')
            
            kaug.options["print_kkt"] = ""
            kaug.solve(self.model_object, tee=True)

            kaug_files = Path('GJH')
            var_index_names = pd.read_csv(self.sol_files['col'], sep=';', header=None)  # dummy sep
            con_index_names = pd.read_csv(self.sol_files['row'], sep=';', header=None)  # dummy sep

            var_index_names = [var_name for var_name in var_index_names[0]]
            con_index_names = [con_name for con_name in con_index_names[0].iloc[:-1]]
            # con_index_number = {v: k for k, v in enumerate(con_index_names)}

            n = len(var_index_names)
            m = len(con_index_names)

            print(f'size: vars: {n}, cons {m}')

            hess_file = kaug_files.joinpath('H_print.txt')
            hess = pd.read_csv(hess_file, delim_whitespace=True, header=None, skipinitialspace=True)
            
            hess.columns = ['irow', 'jcol', 'vals']
            hess.irow -= 1
            hess.jcol -= 1
            # os.unlink(f'{kaug_files}hess_debug.in')

            jac_file = kaug_files.joinpath('A_print.txt')
            jac = pd.read_csv(jac_file, delim_whitespace=True, header=None, skipinitialspace=True)
        
            jac.columns = ['irow', 'jcol', 'vals']
            jac.irow -= 1
            jac.jcol -= 1
            # os.unlink(f'{kaug_files}jacobi_debug.in')

            # try:
            #    duals = read_duals(stub + '.sol')
            # except:
            duals = None

            J = coo_matrix((jac.vals, (jac.jcol, jac.irow)), shape=(m, n))
            Hess_coo = coo_matrix((hess.vals, (hess.irow, hess.jcol)), shape=(n, n))
            H = Hess_coo + triu(Hess_coo, 1).T

        self.delete_sol_files()

        self.kkt_data = {
            'J': J,
            'H': H,
            'var_ind': var_index_names,
            'con_ind': con_index_names,
            'duals': duals,
        }

        return None

    def get_kkt_df(self):
        """Get the Hessian matrix and return it as a data frame

        :return H_df: The hessian matrix as a DataFrame
        :rtype: pandas.DataFrame

        """
        self.get_kkt_info()

        H = self.kkt_data['H']
        var_index_names = self.kkt_data['var_ind']
        con_index_names = self.kkt_data['con_ind']

        H_df = pd.DataFrame(H.todense(), columns=var_index_names, index=var_index_names)

        # See if this works

        # h_con_indx = [k for k in self.model.C.keys()]
        # h_con = [f'Z[{h[0]},{h[1]}]' for h in h_con_indx if h[1] in self.model.measured_data]
        # h_con_indx = [k for k in self.model.U.keys()]
        # h_con += [f'X[{h[0]},{h[1]}]' for h in h_con_indx  if h[1] in self.model.measured_data]

        # col_ind  = [var_ind.loc[var_ind[0] == v].index[0] for v in h_con]
        # Zr = Z[col_ind, :]

        # df_Zr = pd.DataFrame(Zr, index=h_con, columns=[k for k, v in self.model.P.items()])
        # df_Hv = H_df.loc[h_con, h_con]    

        return H_df

    def prep_model_for_k_aug(self):
        """This function prepares the optimization models with required
        suffixes.

        This is important for k_aug

        :return: None

        """
        self.model_object.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        # self.model_object.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        # self.model_object.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        # self.model_object.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        # self.model_object.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
        # self.model_object.red_hessian = Suffix(direction=Suffix.EXPORT)
        # self.model_object.dof_v = Suffix(direction=Suffix.EXPORT)
        # self.model_object.rh_name = Suffix(direction=Suffix.IMPORT)

        count_vars = 1
        for k, v in self.model_object.P.items():
            self.model_object.dof_v[k] = count_vars
            count_vars += 1

        self.model_object.npdp = Suffix(direction=Suffix.EXPORT)

        return None

    def calculate_duals(self):
        """Get duals of the current model

        :return: self.duals
        :rtype: dict
        """

        # For testing - very slow and should not be used!
        if self.kkt_method == 'pynumero':

            nlp = PyomoNLP(self.model_object)
            varList = nlp.get_pyomo_variables()
            conList = nlp.get_pyomo_constraints()
            duals = nlp.get_duals()

            J = nlp.extract_submatrix_jacobian(pyomo_variables=varList, pyomo_constraints=conList)
            H = nlp.extract_submatrix_hessian_lag(pyomo_variables_rows=varList, pyomo_variables_cols=varList)
            J = csc_matrix(J)

            var_index_names = [v.name for v in varList]
            con_index_names = [v.name for v in conList]

            dummy_constraints = [f'{self.global_constraint_name}[{k}]' for k in self.parameter_set]
            jac_row_ind = [con_index_names.index(d) for d in dummy_constraints]
            duals_imp = [duals[i] for i in jac_row_ind]

            self.duals = dict(zip(self.parameter_set, duals_imp))
            if self.verbose:
                print(f'The pynumero results are:')
                print(self.duals)

        else:

            self.duals = {key: self.model_object.dual[getattr(self.model_object, self.global_constraint_name)[key]] for
                          key, val in getattr(self.model_object, self.global_param_name).items()}

            if self.verbose:
                print('The duals are:')
                print(self.duals)

        self.delete_sol_files()

        return self.duals

    def optimize_model(self, d=None):
        """Takes the model object and performs the optimization for a reduced Hessian problem.
        
        :param dict d: The current state of the parameters

        :return: None

        """
        if self.verbose:
            print(f'd: {d}')

        ipopt = SolverFactory('ipopt')
        tmpfile_i = self.get_tmp_file()

        if self.param_con_method == 'global':

            if d is not None:
                param_val_dict = {p: d[i] for i, p in enumerate(self.parameter_set)}
                for k, v in getattr(self.model_object, self.variable_name).items():
                    v.set_value(param_val_dict[k])

            self.add_global_constraints()

            if self.set_param_bounds:
                set_scaled_parameter_bounds(self.model_object,
                                            parameter_set=self.parameter_set,
                                            rho=self.rho,
                                            scaled=self.scaled)

        elif self.param_con_method == 'fixed':

            if hasattr(self.model_object, self.global_constraint_name):
                self.model_object.del_component(self.global_constraint_name)

            delta = 1e-20
            for k, v in getattr(self.model_object, self.variable_name).items():
                if k in self.parameter_set:
                    ub = v.value
                    lb = v.value - delta
                    v.setlb(lb)
                    v.setub(ub)
                    v.unfix()
                else:
                    v.fix()
    
        ipopt.solve(self.model_object,
                    symbolic_solver_labels=True,
                    keepfiles=True,
                    tee=True,
                    logfile=tmpfile_i,
                    )
        # print(self.model_object.P.display())

        # Create the file object so that it can be deleted
        # self.get_file_info()
        
        stub = ipopt._problem_files[0][:-3]

        nl_file = Path(stub + '.nl')
        col_file = Path(stub + '.col')
        row_file = Path(stub + '.row')
        sol_file = Path(stub + '.sol')

        self.sol_files = dict(
            nl=nl_file,
            col=col_file,
            row=row_file,
            sol=sol_file,
        )

        return None

    def calculate_reduced_hessian(self, d=None, optimize=False, return_Z=False):
        """Calculate the reduced Hessian
        
        :param dict d: The current parameter values
        :param bool optimize: Option to optimize the problem (if not done already)
        :param bool return_Z: Option to add the Z_mat to the returned values

        .. note::

            If returning the Z_mat, the returned object will be a tuple of the reduced Hessian and the Z matrix.

        :return: The reduced Hessian and optionally the Z matrix
        :rtype: numpy.ndarray
        
        """
        if optimize:
            self.optimize_model(d)

        self.get_kkt_info()
        H = self.kkt_data['H']
        J = self.kkt_data['J']
        var_ind = self.kkt_data['var_ind']
        con_ind_new = self.kkt_data['con_ind']
        duals = self.kkt_data['duals']

        col_ind = [var_ind.index(f'{self.variable_name}[{v}]') for v in self.parameter_set]
        m, n = J.shape

        if self.param_con_method == 'global':

            dummy_constraints = [f'{self.global_constraint_name}[{k}]' for k in self.parameter_set]
            print(dummy_constraints)
            jac_row_ind = [con_ind_new.index(d) for d in dummy_constraints]
            # duals_imp = [duals[i] for i in jac_row_ind]

            # print(J.shape, len(duals_imp))

            J_c = delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
            row_indexer = SparseRowIndexer(J_c)
            J_f = row_indexer[col_ind]
            J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
            J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
            print(J_f.shape, J_l.shape)

        elif self.param_con_method == 'fixed':
            
            #jac_row_ind = []

            J_c = J.tocsc()  # delete_from_csr(J.tocsr(), row_indices=jac_row_ind).tocsc()
            row_indexer = SparseRowIndexer(J_c.T)
            J_f = row_indexer[col_ind].T
            # J_f = delete_from_csr(J_f.tocsr(), row_indices=jac_row_ind, col_indices=[])
            J_l = delete_from_csr(J_c.tocsr(), col_indices=col_ind)
            # print(J_f, J_l)
            #print(col_ind)
            print(J_f.shape, J_l.shape)

        else:
            None

        r_hess, Z_mat = self._reduced_hessian_matrix(J_f, J_l, H, col_ind)
        if not return_Z:
            return r_hess.todense()
        else:
            return r_hess.todense(), Z_mat

    @staticmethod
    def _reduced_hessian_matrix(F, L, H, col_ind):
        """This calculates the reduced hessian by calculating the null-space based
        on the constraints

        :param csr_matrix F: Rows of the Jacobian related to fixed parameters
        :param csr_matrix L: The remainder of the Jacobian without parameters
        :param csr_matrix H: The sparse Hessian
        :param list col_ind: indices of columns with fixed parameters
        
        :return: sparse version of the reduced Hessian
        :rtype: csr_matrix

        """
        n = H.shape[0]
        n_free = n - F.shape[0]
        X = spsolve(L.tocsc(), -F.tocsc())
        col_ind_left = list(set(range(n)).difference(set(col_ind)))
        col_ind_left.sort()

        Z = np.zeros([n, n_free])
        Z[col_ind, :] = np.eye(n_free)

        if isinstance(X, csc_matrix):
            Z[col_ind_left, :] = X.todense()
        else:
            Z[col_ind_left, :] = X.reshape(-1, 1)

        Z_mat = coo_matrix(np.mat(Z)).tocsr()
        Z_mat_T = coo_matrix(np.mat(Z).T).tocsr()
        Hess = H.tocsr()
        reduced_hessian = Z_mat_T * Hess * Z_mat

        return reduced_hessian, Z_mat

    def add_global_constraints(self):
        """This adds the dummy constraints to the model forcing the local
        parameters to equal the current global parameter values

        :return: None

        """
        if self.parameter_set is None:
            self.parameter_set = [p for p in getattr(self.model_object, self.variable_name)]

        if self.scaled:
            global_param_init = {p: 1 for p in self.parameter_set}
        else:
            global_param_init = {p: getattr(self.model_object, self.variable_name)[p].value for p in self.parameter_set}

        for comp in [self.global_constraint_name, self.global_param_name, self.current_set]:
            if hasattr(self.model_object, comp):
                print(comp)
                
                self.model_object.del_component(comp)

        setattr(self.model_object, self.current_set, Set(initialize=self.parameter_set))

        setattr(self.model_object, self.global_param_name,
                Param(getattr(self.model_object, self.param_set_name),
                      initialize=global_param_init,
                      mutable=True,
                      ))

        def rule_fix_global_parameters(m, k):
            return getattr(m, self.variable_name)[k] - getattr(m, self.global_param_name)[k] == 0

        setattr(self.model_object, self.global_constraint_name,
                Constraint(getattr(self.model_object, self.current_set),
                           rule=rule_fix_global_parameters)
                )

        return None


def read_duals(sol_file):
    """Reads the duals from the sol file after solving the problem

    :param str sol_file: The absolute path to the sol file

    :return list duals: The list of duals values taken from the sol file
    
    """
    sol_file_abs = Path(sol_file)

    duals = []
    with sol_file_abs.open() as f:
        lines = f.readlines()

    lines_found = True
    num_of_vars = int(lines[9])

    for ln, line in enumerate(lines):
        line = line.rstrip('\n')
        line = line.lstrip('\t').lstrip(' ')

        if ln >= 12 and ln <= (num_of_vars + 11):
            duals.append(float(line))

    return duals


def delete_from_csr(mat, row_indices=[], col_indices=[]):
    """
    Remove the rows (denoted by ``row_indices``) and columns (denoted by 
    ``col_indices``) from the CSR sparse matrix ``mat``.
    WARNING: Indices of altered axes are reset in the returned matrix

    :param csr_matrix mat: Sparse matrix to delete rows and cols from
    :param list row_indicies: rows to delete
    :param list col_indicies: cols to delete

    :return csr_matrix mat: The sparse matrix with the rows and cols removed

    """
    if not isinstance(mat, csr_matrix):
        raise ValueError("works only for CSR format -- use .tocsr() first")

    rows = []
    cols = []
    if row_indices:
        rows = list(row_indices)
    if col_indices:
        cols = list(col_indices)

    if len(rows) > 0 and len(cols) > 0:
        row_mask = np.ones(mat.shape[0], dtype=bool)
        row_mask[rows] = False
        col_mask = np.ones(mat.shape[1], dtype=bool)
        col_mask[cols] = False
        return mat[row_mask][:, col_mask]
    elif len(rows) > 0:
        mask = np.ones(mat.shape[0], dtype=bool)
        mask[rows] = False
        return mat[mask]
    elif len(cols) > 0:
        mask = np.ones(mat.shape[1], dtype=bool)
        mask[cols] = False
        return mat[:, mask]
    else:
        return mat


class SparseRowIndexer:
    """Class used to splice sparse matrices"""

    def __init__(self, matrix):
        data = []
        indices = []
        indptr = []

        _class = 'csr'
        # print(f'LOOK HERE: {type(matrix)}')
        if isinstance(matrix, csc_matrix):
            _class = 'csc'
        #     print(_class)

        self._class = _class
        # Iterating over the rows this way is significantly more efficient
        # than matrix[row_index,:] and matrix.getrow(row_index)
        for row_start, row_end in zip(matrix.indptr[:-1], matrix.indptr[1:]):
            data.append(matrix.data[row_start:row_end])
            indices.append(matrix.indices[row_start:row_end])
            indptr.append(row_end - row_start)  # nnz of the row

        self.data = np.array(data, dtype=object)
        self.indices = np.array(indices, dtype=object)
        self.indptr = np.array(indptr)
        self.n_columns = matrix.shape[1]

    def __getitem__(self, row_selector):
        data = np.concatenate(self.data[row_selector])
        indices = np.concatenate(self.indices[row_selector])
        indptr = np.append(0, np.cumsum(self.indptr[row_selector]))

        shape = [indptr.shape[0] - 1, self.n_columns]

        if self._class == 'csr':
            return csr_matrix((data, indices, indptr), shape=shape)
        else:
            return csr_matrix((data, indices, indptr), shape=shape).T.tocsc()



"""
This module contains various tools that are applied to Pyomo models. Most are
used to extract and format data.

"""
# Standared library imports
import itertools

# Third party imports
import pandas as pd
from pyomo.core.base.param import Param
from pyomo.core.base.set import BoundsInitializer, SetProduct
from pyomo.core.base.var import Var
from pyomo.dae.contset import ContinuousSet
from pyomo.dae.diffvar import DerivativeVar

# KIPET library imports
from kipet.general_settings.variable_names import VariableNames


def get_vars(model):
    """Extract the variable information from the Pyomo model
    
    :param ConcreteModel model: A Pyomo model object
    
    :return list model_varialbes: A list of the model's variables
    
    """
    vars_list = []
    model_Var = list(model.component_map(Var))
    model_dVar = list(model.component_map(DerivativeVar))
    vars_list = model_Var + model_dVar

    return vars_list


def get_vars_block(model):
    """Alternative method for getting the model varialbes
    
    :param ConcreteModel model: A Pyomo model object
    
    :return set model_varialbes: A list of the model's variables
    
    """
    model_variables = set()
    for block in model.block_data_objects():
        block_map = block.component_map(Var)
        for name in block_map.keys():
            model_variables.add(name)
        
    return model_variables


def get_params(model):
    """Returns a list of the model parameters
    
    :param ConcreteModel model: A Pyomo model object
    
    :return list param_list: A list of the model's parameters
    
    """
    param_list = list(model.component_map(Param))
    
    return param_list


def get_result_vars(model):
    """Get the Vars and Params needed for the results object
    
    :param ConcreteModel model: A Pyomo model object
    
    :return list result_vars: A list of the models parameters and variables
    
    """
    result_vars = get_vars(model)
    result_vars += get_params(model)
    
    return result_vars


def get_index_sets(model_var_obj):
    """Retuns a list of the index sets for the model variable
    
    :param model_var_obj: A Pyomo model variable
        
        
    :return list index_set: A list of the variable's indecies
    
    """
    index_set = []
    index = model_var_obj.index_set()
    if not isinstance(index, SetProduct):
        index_set.append(index)
    elif isinstance(index, SetProduct): # or isinstance(index, SetProduct):
        index_set.extend(index.subsets())
    else:
        return None
    
    return index_set


def index_set_info(index_list):
    """Returns whether index list contains a continuous set and where the
    index is
    
    :param list index_list: list of indicies produced by get_index_sets
        
    :return tuple cont_set_info: (Bool, index of continuous set)
        
    """
    index_dict = {'cont_set': [],
                  'other_set': [],
                  }

    for i, index_set in enumerate(index_list):
        if isinstance(index_set, ContinuousSet):
            index_dict['cont_set'].append(i)
        else:
            index_dict['other_set'].append(i)
    index_dict['other_set'] = tuple(index_dict['other_set'])
        
    return index_dict


def change_continuous_set(cs, new_bounds):
    """Changes the bounds of the continuous set.
    
    This is primarily used in the fe_factory class to change the bounds on
    the FEs used in simulation.
    
    :param ContinuousSet cs: Continuous set to change bound on
        
    :param tuple new_bounds: New lower and upper bounds for cs
        
    :return: None
    
    """
    cs.clear()
    cs._init_domain._set = None
    cs._init_domain._set = BoundsInitializer(new_bounds)
    domain = cs._init_domain(cs.parent_block(), None)
    cs._domain = domain
    domain.parent_component().construct()
    
    for bnd in cs.domain.bounds():
        if bnd is not None and bnd not in cs:
            cs.add(bnd)
    cs._fe = sorted(cs)
    
    return None


def convert(var):
    """Load variables from the pyomo model into various formats.
    
    If to_load is None all of the model variables will be considered.
    
    :param instance: Pyomo ConcreteModel instance
    :param name str: The name of the model variable
    
    :return var_data: This is the variable and its type depends on the
        dimensions of the data (float, Series, DataFrame)
    
    """ 
    if var.dim()==0:
        var_data = var.value
    elif var.dim()==1:
        var_data = pd.Series(var.extract_values())
    elif var.dim()==2:
        d = var.extract_values()
        keys = d.keys()
        if keys:
            var_data = _df_from_pyomo_data(var)
        else:
            var_data = pd.DataFrame(data=[],
                                    columns = [],
                                    index=[])   
     
    else:
        var_data = _prepare_data_for_init(var)
        
    return var_data
        
        
def _prepare_data_for_init(var):
        """Convert results dict into DataFrame for initialization.
        
        This is used for data with more than two dimensions. It works by
            grouping the remaining dimensions as tuples in the columns of the
            DataFrame.
            
        :param model: The Pyomo model object
        :param var: The target variable in model
        
        :return pandas.DataFrame df: The extracted variable data as a two
            dimensional DataFrame
            
        """
        if len(var) == 0:
            return None
            
        model = var.model()
        index_sets = get_index_sets(var)
        index_dict = index_set_info(index_sets)
        time_set = index_sets[index_dict['cont_set'][0]].name
        component_indecies = index_dict['other_set']
        component_sets = [index_sets[i].name for i in component_indecies]
        index = getattr(model, time_set).value_list
        columns = list(itertools.product(*[getattr(model, comp_list).value_list for comp_list in component_sets]))
        df = pd.DataFrame(data=None, index=index, columns=columns)

        for i in index:
            for j in columns:
                jl = list(j)
                jl.insert(index_dict['cont_set'][0], i)
                df.loc[i,j] = var[tuple(jl)].value
                
        return df
    
    
def _df_from_pyomo_data(var):
    """Takes a variable object from a Pyomo model and returns a pandas
    DataFrame instance of the data it contains.
    
    The returned DataFrame has time as the index.
    
    :param var: An instance of a Pyomo variable
    
    :return pandas.DataFrame dfs: A DataFrame containing the data of the 
        variable.
    
    """
    val = []
    ix = []
    for index in var:
        ix.append(index)
        try:
            val_raw = var[index].value
        except:
            val_raw = var[index]
            
        if val_raw is None:
            val_raw = 0
        val.append(val_raw)
    
    a = pd.Series(index=ix, data=val)
    dfs = pd.DataFrame(a)
    index = pd.MultiIndex.from_tuples(dfs.index)
   
    dfs = dfs.reindex(index)
    dfs = dfs.unstack()
    dfs.columns = [v[1] for v in dfs.columns]

    return dfs


def model_info(model):
    """This method provides a dict of model attributes to be used in various
    classes.
    
    .. note ::
        
        This may be moved to the TemplateBuilder since this may be redundant
    
    :param ConcreteModel model: A Pyomo model
    
    :return: Useful model attributes
    :rtype: dict
    
    """
    __var = VariableNames()
    model_attrs = {}
    
    model_attrs['mixture_components'] = list(model.mixture_components)
    model_attrs['complementary_states'] = list(model.complementary_states)
    model_attrs['algebraics'] = list(model.algebraics)
    model_attrs['n_components'] = len(model_attrs['mixture_components'])
    model_attrs['n_algebraics'] = len(model_attrs['algebraics'])
    model_attrs['n_complementary_states'] = len(model_attrs['complementary_states'])
   
    model_attrs['known_absorbance_data'] = None
    
    model_attrs['non_absorbing'] = None
    if hasattr(model, 'non_absorbing'):
        model_attrs['non_absorbing'] = list(model.non_absorbing)
        
    model_attrs['known_absorbance'] = None
    if hasattr(model, 'known_absorbance'):
        model_attrs['known_absorbance'] = list(model.known_absorbance)

    if hasattr(model, 'abs_components'):
        model_attrs['abs_components'] = list(model.abs_components.keys())
        model_attrs['nabs_components'] = len(model_attrs['abs_components'])

    model_attrs['huplc_absorbing'] = None
    if hasattr(model, 'huplc_absorbing'):
        model_attrs['huplc_absorbing'] =  list(model.huplc_absorbing)
    
    if hasattr(model, 'huplcabs_components'):
        model_attrs['huplcabs_components'] = list(model.huplcabs_components)
        model_attrs['nhuplcabs_components'] = len(model_attrs['huplcabs_components'])
    
    model_attrs['ipopt_scaled'] = False
    model_attrs['spectra_given'] = hasattr(model, __var.spectra_data)
    model_attrs['concentration_given'] = hasattr(model, __var.concentration_measured) or hasattr(model, __var.user_defined) or hasattr(model, __var.state)
    model_attrs['conplementary_states_given'] = hasattr(model, __var.state)
    model_attrs['absorption_given'] = hasattr(model, __var.spectra_species)
    model_attrs['huplc_given'] = hasattr(model, 'Chat')
    model_attrs['smoothparam_given'] = hasattr(model, __var.smooth_parameter)
                
    return model_attrs



"""
Convenience class to hold the results from the Pyomo models after simulation
and parameter fitting.

"""
# Standard library imports

# Third party imports
import numpy as np

# Kipet library imports
from kipet.model_tools.pyomo_model_tools import convert, get_vars

# This needs deletion at some point!
result_vars = ['Z', 'C', 'Cm', 'K', 'S', 'X', 'dZdt', 'dXdt', 'P', 'Pinit', 'sigma_sq', 'estimable_parameters', 'Y', 'UD', 'step']


class ResultsObject(object):
    """Container for all of the results from the Pyomo model"""
    
    def __init__(self):
        """
        A class to store simulation and optimization results.
        
        The ResultsObject instance takes no initial parameters. This will most
        likely change in the near future.
        
        """
        self.parameter_covariance = None

    def __str__(self):
        string = "\nRESULTS\n"
        
        # results_vars = get_results_vars()
        
        for var in result_vars:
            if hasattr(self, var) and getattr(self, var) is not None:
                var_str = var
                if var == 'sigma_sq':
                    var_str = 'Sigmas2'
                string += f'{var_str}:\n {getattr(self, var)}\n\n'
        
        return string
    
    def __repr__(self):
        
        return self.__str__()

    def _confidence_interval_display(self, interval=0.95):
        """
        Function to display calculated confidence intervals

        :param dict variances: The component variances

        :return: None

        """
        margin = 15
        if self.parameter_covariance is not None:
            deviations = self.deviations(interval)
            
            print(f'\n# Parameter Values with confidence: ({int(interval*100)}%)')
            for k, p in self.P.items():
                if k in self.variances:
                    print(f'{k.rjust(margin)} = {p:0.4e} +/- {deviations[k]:0.4e}') 
            if hasattr(self, 'Pinit'):
                for k, p in self.Pinit.items():
                    print(f'{k.rjust(margin)} = {p:0.4e} +/- {deviations[k]:0.4e}') 
            if hasattr(self, 'time_step_change'):
                for k, p in self.time_step_change.items():
                    if k in self.variances:
                        print(f'{k.rjust(margin)} = {p:0.4e} +/- {deviations[k]:0.4e}') 
        else:
            print(f'\n# Parameter Values')
            for k, p in self.P.items():
                print(f'{k.rjust(margin)} = {p:0.4e}')
            if hasattr(self, 'Pinit'):
                for k, p in self.Pinit.items():
                    print(f'{k.rjust(margin)} = {p:0.4e}') 
            if hasattr(self, 'time_step_change'):
                for k, p in self.time_step_change.items():
                    print(f'{k.rjust(margin)} = {p:0.4e}') 
            
        return None

    def deviations(self, interval=0.95):
        """Calculates bounds for the parameters based on STD and condifence
        interval provided
        
        :param float interval: The confidence interval
        
        """
        if not hasattr(self, 'parameter_covariance') or self.parameter_covariance is None:
            return None
        
        import scipy.stats as st
        dev = st.norm.ppf(1-(1-interval)/2)
        
        deviations = {k: dev*v**0.5 for k, v in self.variances.items()}
        
        return deviations

    def compute_var_norm(self, variable_name, norm_type=np.inf):
        var = getattr(self, variable_name)
        var_array = np.array(var)
        return np.linalg.norm(var_array,norm_type)

    def load_from_pyomo_model(self, model, to_load=None):
        """Load variables from the pyomo model into various formats.
        
        This will set the attribute of all the model variables in a specific 
        format depending on the dimensionality of the variable into the 
        ResultsObject.
        
        :param ConcreteModel model: Model of the reaction system
        
        :return: None
        
        """
        variables_to_load = get_vars(model)
    
        for name in variables_to_load:
    
            if name == 'init_conditions':
                continue
            
            var = getattr(model, name)
            var_data = convert(var)
            setattr(self, name, var_data)
                

    def parameters(self, interval=0.95):
        """Returns the parameter dictionary
        
        :return: The parameter values from the model
        :rtype: dict
        
        """
        return self._confidence_interval_display(interval)
            

    def show_parameters(self, interval=0.95):
        """Displays the parameter values in a convenient manner
        
        :return: None
        
        """
        return self._confidence_interval_display(interval)
            
    @property
    def variances(self):
        """Displays the variances of the parameters in a convenient manner
        
        :return: None
        
        """
        if not hasattr(self, 'parameter_covariance') or self.parameter_covariance is None:
            return None
        
        var = dict(zip(self.parameter_covariance.columns, 
                       np.diag(self.parameter_covariance.values)
            ))
        
        return var


"""
Primary class for performing the parameter fitting in KIPET
"""
# Standard library imports
import copy
import time

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import (
    Constraint, 
    ConstraintList, 
    Objective,
    SolverFactory, 
    TerminationCondition, 
    value, 
    Var)

# KIPET library imports
from kipet.model_components.objectives import (
    comp_objective,
    conc_objective)
from kipet.estimator_tools.pyomo_simulator import PyomoSimulator
from kipet.estimator_tools.reduced_hessian_methods import (
    covariance_k_aug,
    covariance_sipopt, 
    define_free_parameters)
from kipet.estimator_tools.results_object import ResultsObject
from kipet.mixins.parameter_estimator_mixins import PEMixins
from kipet.model_tools.pyomo_model_tools import convert
from kipet.general_settings.variable_names import VariableNames


class ParameterEstimator(PEMixins, PyomoSimulator):

    """Optimizer for parameter estimation"""

    def __init__(self, model):
        super(ParameterEstimator, self).__init__(model)

        self.__var = VariableNames()

        self.hessian = None
        self.cov = None
        self._estimability = False
        self._idx_to_variable = dict()
        self.model_variance = True
        self.termination_condition = None
        
        # This should be a subclass or a mixin
        self.G_contribution = None
        self.unwanted_G = False
        self.time_variant_G = False
        self.time_invariant_G = False
        self.time_invariant_G_decompose = False
        self.time_invariant_G_no_decompose = False
        
        # Parameter name list
        self.param_names_full = define_free_parameters(self.model, kind='full')
        self.param_names = define_free_parameters(self.model, kind='simple')

        # for new huplc structure (CS):
        if self._huplc_given:
            self._list_huplcabs = self._huplc_absorbing
            # self._n_huplc = len(self._list_huplcabs)

        self.n_val = self._n_components
        
        self.cov = None
            
    def run_opt(self, solver, **kwds):

        """ Solves parameter estimation problem.

        :param str solver: The solver used to solve the NLP
        :param dict kwds: The dictionary of options passed from the ReactionModel

        :Keyword Args:
            - solver (str): name of the nonlinear solver to used
            - solver_opts (dict, optional): options passed to the nonlinear solver
            - variances (dict or float, optional): map of component name to noise variance. The
              map also contains the device noise variance. If not float then we only use device variance
              and ignore model variance.
            - tee (bool,optional): flag to tell the optimizer whether to stream output
              to the terminal or not.
            - with_d_vars (bool,optional): flag to the optimizer whether to add
            - variables and constraints for D_bar(i,j).
            - report_time (bool, optional): flag as to whether to time the parameter estimation or not.
            - estimability (bool, optional): flag to tell the model whether it is
              being used by the estimability analysis and therefore will need to return the
              hessian for analysis.
            - model_variance (bool, optional): Default is True. Flag to tell whether we are only
              considering the variance in the device, or also model noise as well.
            - model_variance (bool, optional): Default is True. Flag to tell whether we are only
              considering the variance in the device, or also model noise as well.

        :return: The results from the parameter fitting
        :rtype: ResultsObject

        """
        run_opt_kwargs = copy.copy(kwds)
        
        solver_opts = kwds.pop('solver_opts', dict())
        variances = kwds.pop('variances', dict())
        tee = kwds.pop('tee', False)
        with_d_vars = kwds.pop('with_d_vars', False)
        covariance = kwds.pop('covariance', None)
 
        estimability = kwds.pop('estimability', False)
        report_time = kwds.pop('report_time', False)
        model_variance = kwds.pop('model_variance', True)

        inputs_sub = kwds.pop("inputs_sub", None)
        trajectories = kwds.pop("trajectories", None)
        fixedtraj = kwds.pop("fixedtraj", False)
        fixedy = kwds.pop("fixedy", False)
        yfix = kwds.pop("yfix", None)
        yfixtraj = kwds.pop("yfixtraj", None)

        jump = kwds.pop("jump", False)
        
        self.confidence = kwds.pop('confidence', None)
        if self.confidence is None:
            self.confidence = 0.95
        
        G_contribution = kwds.pop('G_contribution', None)
        St = kwds.pop('St', dict())
        Z_in = kwds.pop('Z_in', dict())

        self.solver = solver
        self.covariance_method = covariance
        self.model_variance = model_variance
        self._estimability = estimability

        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before initializing')

        self.G_contribution = G_contribution
       
        if report_time:
            start = time.time()
        
        if self.G_contribution == 'time_invariant_G':
            self.decompose_G_test(St, Z_in)
        self.g_handling_status_messages()

        # Move?
        if self.covariance_method == 'ipopt_sens':
            if not 'compute_red_hessian' in solver_opts.keys():
                solver_opts['compute_red_hessian'] = 'yes'
                
        self.solver_opts = solver_opts
            
        if inputs_sub is not None:
            from kipet.estimator_tools.additional_inputs import add_inputs
            
            add_kwargs = dict(
                fixedtraj = fixedtraj,
                fixedy = fixedy, 
                inputs_sub = inputs_sub,
                yfix = yfix,
                yfixtraj = yfixtraj,
                trajectories = trajectories,
            )
            add_inputs(self, add_kwargs)
        
        if jump:
            from kipet.estimator_tools.jumps_method import set_up_jumps
            set_up_jumps(self.model, run_opt_kwargs)
            
        # # I am not sure if this is still needed...
        # active_objectives = [o for o in self.model.component_map(Objective, active=True)]        
        # if active_objectives:
        #     print(
        #         "WARNING: The model has an active objective. Running optimization with models objective.\n"
        #         " To solve optimization with default objective (Weifengs) deactivate all objectives in the model.")
        #     solver_results = opt.solve(self.model, tee=tee)

        #if self._spectra_given:
        self.objective_value = self._solve_model(
            variances,
            tee=tee,
            covariance=covariance,
            with_d_vars=with_d_vars,
            **kwds)

        if report_time:
            end = time.time()
            print("Total execution time in seconds for variance estimation:", end - start)

        return self._get_results()

    def _get_results(self):
        """Removed results unit from function

        :return: The formatted results
        :rtype: ResultsObject

        """
        results = ResultsObject()
        results.objective = self.objective_value
        results.parameter_covariance = self.cov
        results.load_from_pyomo_model(self.model)
        results.show_parameters(self.confidence)

        if self._spectra_given:
            from kipet.calculation_tools.beer_lambert import D_from_SC
            D_from_SC(self.model, results)

        if hasattr(self.model, self.__var.model_parameter_scaled): 
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value*getattr(self.model, self.__var.model_parameter_scaled)[name].value for name in self.model.parameter_names})
        else:
            setattr(results, self.__var.model_parameter, {name: getattr(self.model, self.__var.model_parameter)[name].value for name in self.model.parameter_names})

        if self.termination_condition!=None and self.termination_condition!=TerminationCondition.optimal:
            raise Exception("The current iteration was unsuccessful.")
        else:
            if self._estimability == True:
                return self.hessian, results
            else:
                return results

        return results
            
    def _solve_model(self, sigma_sq, **kwds):
        """Main function to getup the objective and perform the parameter estimation 

           This method is not intended to be used by users directly

        :param dict sigma_sq: variances
        :param SolverFactory optimizer: Pyomo Solver factory object

        :Keyword Args:

            - with_d_vars (bool,optional): flag to the optimizer whether to add
              variables and constraints for D_bar(i,j)

        :return: None

        """
        with_d_vars = kwds.pop('with_d_vars', False)
     
        # These are not being used
        penaltyparam = kwds.pop('penaltyparam', False)
        penaltyparamcon = kwds.pop('penaltyparamcon', False) #added for optional penalty term related to constraint CS
        ppenalty_dict = kwds.pop('ppenalty_dict', None)
        ppenalty_weights = kwds.pop('ppenalty_weights', None)
        
        model = self.model
        model.objective = Objective(expr=0)
        
        if self._spectra_given:

            all_sigma_specified = True
    
            if isinstance(sigma_sq, dict): 
                keys = sigma_sq.keys()
                for k in self.comps['unknown_absorbance']:
                    if k not in keys:
                        all_sigma_specified = False
                        sigma_sq[k] = max(sigma_sq.values())
    
                if not 'device' in sigma_sq.keys():
                    all_sigma_specified = False
                    sigma_sq['device'] = 1.0
    
            elif isinstance(sigma_sq, float):
                sigma_dev = sigma_sq
                sigma_sq = dict()
                sigma_sq['device'] = sigma_dev
            
            if self.time_invariant_G_no_decompose:
                for i in model.times_spectral:
                    model.qr[i] = 1.0
                model.qr.fix()
    
            def _qr_end_constraint(model):
                return model.qr[model.times_spectral[-1]] == 1.0
            
            if self.G_contribution == 'time_variant_G':
                model.qr_end_cons = Constraint(rule = _qr_end_constraint)
            
            
            if with_d_vars and self.model_variance:
                model.D_bar = Var(model.times_spectral, model.meas_lambdas)
                
                def rule_D_bar(model, t, l):
                    if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                        return model.D_bar[t, l] == sum(
                            getattr(model, 'C')[t, k] * model.S[l, k] for k in self.comps['unknown_absorbance'] if k not in model.solid_spec_arg1)
                    else:
                        return model.D_bar[t, l] == sum(getattr(model, 'C')[t, k] * model.S[l, k] for k in self.comps['unknown_absorbance'])
    
                model.D_bar_constraint = Constraint(model.times_spectral,
                                                    model.meas_lambdas,
                                                    rule=rule_D_bar)
    
            if self.model_variance:
            
                # D_bar - Beer Lambert's Law
                self._primary_D_term(model, with_d_vars, self.comps['unknown_absorbance'], sigma_sq)
                # Model fit
                self._concentration_spectra_term(model, sigma_sq)     
                # L2 penalty
                self._l2_term(model, penaltyparam, ppenalty_dict, ppenalty_weights)
                # HUPLC
                if hasattr(model, 'huplc_absorbing'):
                    self._huplc_obj_term(model, sigma_sq)
                    self._penalty_term(model, penaltyparamcon, rho=1e-1)
                    
            else: # device only
                if hasattr(model, 'huplc_absorbing'):
                    component_set = self.comps['absorbing']
                    self._huplc_obj_term(model, sigma_sq)
                else:
                    component_set = self.comps['unknown_absorbance']
                
                self._primary_D_term(model, with_d_vars, component_set, sigma_sq, device=True)
         
        elif self._concentration_given: 
            
            self._concentration_term(model, sigma_sq, 'concentration')
            self._state_term(model)
            self._custom_term(model)
            self._penalty_term(model, penaltyparamcon)
         
        # # Break up the objective
        # custom_weight = 1
        # if hasattr(model, 'custom_obj'):
        #     model.objective.expr += custom_weight*(model.custom_obj)

        # HUPLC shows up here too - if needed in a future version, this can be reimplemented
        # for k in self.comps['all']:
        #     if hasattr(self, 'huplc_absorbing') and k not in keys:
        #         for ki in self.solid_spec_args1:
        #             for key in keys:
        #                 if key == ki:
        #                     sigma_sq[ki] = sigma_sq[key]

        #     elif hasattr(self, 'huplc_absorbing') == False and k not in keys:
        #         all_sigma_specified = False
        #         sigma_sq[k] = max(sigma_sq.values())

        # if self._huplc_given:  # added for new huplc structure CS
        #     weights = kwds.pop('weights', [1.0, 1.0, 1.0])
        # else:
        #     weights = kwds.pop('weights', [1.0, 1.0])

        obj_val = self.optimize(model, sigma_sq)
        
        return obj_val

    @staticmethod
    def _concentration_term(model, sigma_sq, source):
        """Sets up the LSE terms for the concentration in the objective
        
        :param ConcreteModel model: The current model
        :param dict sigma_sq: The component variances
        :param str source: concentration or specta
        
        :return: None
        
        """

        obj=0
        obj += conc_objective(model, variance=sigma_sq, source=source)  
        model.objective.expr += obj
    
        return None
    
    @staticmethod
    def _state_term(model):
        """Sets up the LSE terms for the states in the objective
        
        :param ConcreteModel model: The current model
        
        :return: None
        
        """
        obj = comp_objective(model)  
        model.objective.expr += obj
    
    @staticmethod
    def _custom_term(model):
        """Sets up the LSE terms for the custom objective
        
        :param ConcreteModel model: The current model
        
        :return: None
        
        """
        if hasattr(model, 'custom_obj'):
            model.objective.expr += model.custom_obj
    
        return None
    
    def _concentration_spectra_term(self, model, sigma_sq):
        """Sets up the LSE terms for the concentration (from spectra) in the objective
        
        :param ConcreteModel model: The current model
        :param dict sigma_sq: The component variances

        :return: None
        
        """
        obj = 0
        for t in model.times_spectral:
            obj += sum((model.C[t, k] - model.Z[t, k]) ** 2 / sigma_sq[k] for k in self.comps['unknown_absorbance'])
            
        model.objective.expr += obj
        
        return None
    
    @staticmethod
    def _penalty_term(model, penaltyparamcon, rho=100):
        """Sets up the penalty terms in the objective
        
        :param ConcreteModel model: The current model
        :param bool penaltyparacon: option whether penalties are used
        :param float rho: weighting for the penalties
        
        :return: None
        
        """
        if penaltyparamcon == True:
            sum_penalty = 0.0
            for t in model.allmeas_times:
                sum_penalty += model.Y[t, 'npen']

            model.objective.expr += rho * sum_penalty
            
        return None
    
    @staticmethod
    def _l2_term(model, penaltyparam, ppenalty_dict, ppenalty_weights):
        """Sets up the L2 penatly terms in the objective
        
        :param ConcreteModel model: The current model
        :param dict penatlyparam: option if penalty params are used
        :param dict ppenalty_dict: set-up info for the penalty parameters
        :param dict ppenalty_weights: weights associated with penalties
        
        :return: None
        
        """
        terms = 0
        # L2 penalty term to objective penalizing values that deviate from values defined in ppenalty_dict:
        if penaltyparam==True:
            if ppenalty_weights is None:
                terms = 0.0
                for k in model.P.keys():
                    if k in ppenalty_dict.keys():
                        terms += (model.P[k] - ppenalty_dict[k]) ** 2
            else:
                if len(ppenalty_dict)!=len(ppenalty_weights):
                    raise RuntimeError(
                        'For every penalty term a weight must be defined.')
                if ppenalty_dict.keys()!=ppenalty_weights.keys():
                    raise RuntimeError(
                        'Check the parameter names in ppenalty_weights and ppenalty_dict again. They must match.')
                else:
                    for k in model.P.keys():
                        if k in ppenalty_dict.keys():
                                terms += ppenalty_weights[k] * (model.P[k] - ppenalty_dict[k]) ** 2
                                
        model.objective.expr += terms
    
    def _primary_D_term(self, model, with_d_vars, component_set, sigma_sq, device=False):
        """Build the objective expression for D_bar (primary objective term in KIPET)

        :param ConcreteModel model: The model for use in parameter fitting
        :param bool with_d_vars: For spectral problems
        :param list component_set: A list of the components
        :param dict sigma_sq: A dictionary of component variances
        :param bool device: Indicates if the device variance is used

        :return expr: The objective expression
        :rtype: expression

        """
        expr = 0
        for t in model.times_spectral:
            for l in model.meas_lambdas:
                
                if with_d_vars:
                    D_bar = model.D_bar[t, l]
                else:
                    if device:    
                        if hasattr(model, 'huplc_absorbing'):
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                        else:
                            D_bar = sum(model.Z[t, k] * model.S[l, k] for k in component_set)    
                    else:
                        if hasattr(model, 'huplc_absorbing') and hasattr(model, 'solid_spec_arg1'):
                            D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set if k not in model.solid_spec_arg1)
                        else:
                            D_bar = sum(model.C[t, k] * model.S[l, k] for k in component_set)
                
                if self.G_contribution == 'time_variant_G':
                    expr += (model.D[t, l] - D_bar - model.qr[t]*model.g[l]) ** 2 / (sigma_sq['device'])
                elif self.time_invariant_G_no_decompose:
                    expr += (model.D[t, l] - D_bar - model.g[l]) ** 2 / (sigma_sq['device'])
                else:
                    expr += (model.D[t, l] - D_bar) ** 2 / (sigma_sq['device'])
                   
        model.objective.expr += expr
            
        return None
    
    
    def _regular_optimization(self, model):
        """Optimize without the covariance.
        
        This is used for setting up the solver factory with ipopt for models
        where the covariance is not needed. This is also used for the first
        step in solving models with k_aug.
        
        """
        optimizer = SolverFactory(self.solver)
        for key, val in self.solver_opts.items():
            optimizer.options[key] = val
        solver_results = optimizer.solve(model, tee=False, symbolic_solver_labels=True)
        self._termination_problems(solver_results, optimizer)
    
    
    def optimize(self, model, sigma_sq=None):
        """Handler for optimization using covariance or not"""
        
        if self.covariance_method in ['k_aug', 'ipopt_sens']:
            optimizer = SolverFactory(self.covariance_method)
            for key, val in self.solver_opts.items():
                optimizer.options[key] = val
            self.covariance(optimizer, sigma_sq) 
        else:
            self._regular_optimization(model)
            
        obj_val = model.objective.expr()
       
        model.del_component('objective')
        if hasattr(model, 'D_bar'):
            model.del_component('D_bar')
        if hasattr(model,' D_bar_constraint'):
            model.del_component('D_bar_constraint')
            
        return obj_val
    
    
    def covariance(self, optimizer=None, sigma_sq=None):
        """This consolidates the covariances methods into a convenient function
        
        :param SolverFactory optimizer: The solver (ipopt) used in k_aug
        :param dict sigma_sq: The variances used in covariances calculations
        
        :return: None
        
        """
        if self.covariance_method == 'ipopt_sens':
            self.inv_hessian, self.inv_hessian_reduced = covariance_sipopt(self.model, optimizer, self.comps['unknown_absorbance'], self.param_names_full)
            
        elif self.covariance_method == 'k_aug':
            self.inv_hessian, self.inv_hessian_reduced = covariance_k_aug(self.model, None, self.comps['unknown_absorbance'], self.param_names_full)
            
        if hasattr(self.model, 'C'):
            from kipet.estimator_tools.reduced_hessian_methods import compute_covariance
            models_dict = {'reaction_model': self.model}
            free_params = len(self.param_names)
            variances = {'reaction_model': sigma_sq}
            self.covariance_parameters = compute_covariance(models_dict, self.inv_hessian_reduced, free_params, variances)
        else:
            self.covariance_parameters = self.inv_hessian_reduced
            
        self.cov = pd.DataFrame(self.covariance_parameters, index=self.param_names, columns=self.param_names)  
    
        return None
    
    
    def _termination_problems(self, solver_results, optimizer):
        """Checks the termination conditions and will try once again if it fails

        :param ResultsObject solver_results: The results from the parameter fittings
        :param SolverFactory optimizer: The solver factory being used

        :return: None

        """
        self.termination_condition = solver_results.solver.termination_condition
        if self.termination_condition != TerminationCondition.optimal:
            print("WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
            optimizer.options["OF_start_with_resto"] = 'yes'
            solver_results = optimizer.solve(self.model, tee=False, symbolic_solver_labels=True)
            self.termination_condition = solver_results.solver.termination_condition
            if self.termination_condition != TerminationCondition.optimal:
                print(
                    "WARNING: The solution of the iteration was unsuccessful. The problem is solved with additional solver options.")
                optimizer.options["OF_start_with_resto"] = 'no'
                # optimizer.options["OF_bound_push"] = 1E-02
                optimizer.options["OF_bound_relax_factor"] = 1E-05
                solver_results = optimizer.solve(self.model, tee=False, symbolic_solver_labels=True)
                self.termination_condition = solver_results.solver.termination_condition
                # options["OF_bound_relax_factor"] = 1E-08
                if self.termination_condition != TerminationCondition.optimal:
                        print("The current iteration was unsuccessful.")
                        
        return None
    
        
    def _huplc_obj_term(self, model, sigma_sq):
        """Adds the HUPLC term to the objective

        :param ConcreteModel m: The model used in parameter fitting
        :param dict sigma_sq: The dict of variances

        :return expressions third_term: The HUPLC objective expression

        """
        def rule_Dhat_bar(m, t, l):
            list_huplcabs = [k for k in m.huplc_absorbing.value]
            return m.Dhat_bar[t, l] == (m.Z[t, l] + m.solidvol[t, l]) / (
                sum(m.Z[t, j] + m.solidvol[t, j] for j in list_huplcabs))

        third_term = 0.0
        
        # Arrange the HUPLC device variance
        if not 'device-huplc' in sigma_sq.keys():
            sigma_sq['device-huplc'] = 1.0
            
        if hasattr(model, 'solid_spec_arg1') and hasattr(model, 'solid_spec_arg2'):
            solidvol_dict = dict()
            model.add_component('cons_solidvol', ConstraintList())
            new_consolidvol = getattr(model, 'cons_solidvol')
            model.add_component('solidvol', Var(model.huplctime, model.huplc_absorbing.value, initialize=solidvol_dict))

            for k in model.solid_spec_arg1:
                for j in model.algebraics:
                    for time in model.huplctime:
                        if j == k:
                            for l in model.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in model.solid_spec_arg2 and strsolidspec in str(
                                        model.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valY = value(model.Y[time, k]) / value(model.vol)
                                    if valY <= 0:
                                        new_consolidvol.add(model.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(model.solidvol[time, l] == model.Y[time, k] / model.vol)
                                else:
                                    new_consolidvol.add(model.solidvol[time, l] == 0.0)
                for jk in self.comps['all']:
                    for time in model.huplctime:
                        if jk == k:
                            for l in model.huplc_absorbing.value:
                                strsolidspec = "\'" + str(k) + "\'" + ', ' + "\'" + str(
                                    l) + "\'"  # pair of absorbing solid and liquid
                                if l in model.solid_spec_arg2 and strsolidspec in str(
                                        model.solid_spec.keys()):  #check whether pair of solid and liquid in keys and whether liquid in huplcabs species
                                    valZ = value(model.Z[time, k]) / value(model.vol)
                                    if valZ <= 0:
                                        new_consolidvol.add(model.solidvol[time, l] == 0.0)
                                    else:
                                        new_consolidvol.add(model.solidvol[time, l] == model.Z[time, k] / model.vol)
                                else:
                                    new_consolidvol.add(model.solidvol[time, l] == 0.0)

            model.Dhat_bar = Var(model.huplcmeas_times,
                             model.huplc_absorbing)

            model.Dhat_bar_constraint = Constraint(model.huplcmeas_times,
                                               model.huplc_absorbing,
                                               rule=rule_Dhat_bar)

            for t in model.huplcmeas_times:
                list_huplcabs = [k for k in model.huplc_absorbing.value]
                for k in list_huplcabs:
                    third_term += (model.Dhat[t, k] - model.Dhat_bar[t, k]) ** 2 / sigma_sq['device-huplc']

        model.objective.expr += third_term

        return None

    def run_param_est_with_subset_lambdas(self, builder_clone, end_time, subset, nfe, ncp, sigmas, solver='ipopt'):
        """ Performs the parameter estimation with a specific subset of wavelengths.
            At the moment, this is performed as a totally new Pyomo model, based on the
            original estimation. Initialization strategies for this will be needed.

        :param TemplateBuilder builder_clone: Template builder class of complete model without the data added yet
        :param float end_time: the end time for the data and simulation
        :param list subset: list of selected wavelengths
        :param int nfe: number of finite elements
        :param int ncp: number of collocation points
        :param dict sigmas: dictionary containing the variances, as used in the ParameterEstimator class

        :return ResultsObject results: The solved pyomo model results

        """
        if not isinstance(subset, (list, dict)):
            raise RuntimeError("subset must be of type list or dict!")

        if isinstance(subset, dict):
            lists1 = sorted(subset.items())
            x1, y1 = zip(*lists1)
            subset = list(x1)
        
        # This is the filter for creating the new data subset
        old_D = convert(self.model.D)
        new_D = old_D.loc[:, subset] 
        
        
        print(end_time, new_D)
        # Now that we have a new DataFrame, we need to build the entire problem from this
        # An entire new ParameterEstimation problem should be set up, on the outside of
        # this function and class structure, from the model already developed by the user.
        new_template = construct_model_from_reduced_set(builder_clone, end_time, new_D)
        # need to put in an optional running of the variance estimator for the new
        # parameter estiamtion run, or just use the previous full model run to initialize...

        results, lof = run_param_est(new_template, nfe, ncp, sigmas, solver=solver)

        return results

    def run_lof_analysis(self, builder_before_data, end_time, correlations, lof_full_model, nfe, ncp, sigmas,
                         step_size=0.2, search_range=(0, 1)):
        """ Runs the lack of fit minimization problem used in the Michael's Reaction paper
        from Chen et al. (submitted). To use this function, the full parameter estimation
        problem should be solved first and the correlations for wavelngths from this optimization
        need to be supplied to the function as an option.


        :param TemplateBuilder builder_before_data: Template builder class of complete model without the data added yet
        :param int end_time: the end time for the data and simulation
        :param dict correlations: dictionary containing the wavelengths and their correlations to the concentration profiles
        :param int lof_full_model: the value of the lack of fit of the full model (with all wavelengths)
        :param int nfe: number of finite elements
        :param int ncp: number of collocation points
        :param dict sigmas: dictionary containing the variances, as used in the ParameterEstimator class
        :param float step_size: The spacing used in correlation thresholds
        :param tuple search_range: correlation bounds within to search

        :return: None

        """
        if not isinstance(step_size, float):
            raise RuntimeError("step_size must be a float between 0 and 1")
        elif step_size >= 1 or step_size <= 0:
            return RuntimeError("step_size must be a float between 0 and 1")

        if not isinstance(search_range, tuple):
            raise RuntimeError("search range must be a tuple")
        elif search_range[0] < 0 or search_range[0] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range lower value must be between 0 and 1 and must be type float")
        elif search_range[1] < 0 or search_range[1] > 1 and not (
                isinstance(search_range, float) or isinstance(search_range, int)):
            raise RuntimeError("search range upper value must be between 0 and 1 and must be type float")
        elif search_range[1] <= search_range[0]:
            raise RuntimeError("search_range[1] must be bigger than search_range[0]!")
        # firstly we will run the initial search from at increments of 20 % for the correlations
        # we already have lof(0) so we want 10,30,50,70, 90.
        
        count = 0
        filt = 0.0
        initial_solutions = list()
        initial_solutions.append((0, lof_full_model))
        while filt < search_range[1]:
            filt += step_size
            if filt > search_range[1]:
                break
            elif filt == 1:
                break
            new_subs = wavelength_subset_selection(correlations=correlations, n=filt)
            lists1 = sorted(new_subs.items())
            x1, y1 = zip(*lists1)
            x = list(x1)

            old_D = convert(self.model.D)
            new_D = old_D.loc[:, new_subs]

            # opt_model, nfe, ncp = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # Now that we have a new DataFrame, we need to build the entire problem from this
            # An entire new ParameterEstimation problem should be set up, on the outside of
            # this function and class structure, from the model already developed by the user.
            new_template = construct_model_from_reduced_set(builder_before_data, end_time, new_D)
            # need to put in an optional running of the variance estimator for the new
            # parameter estimation run, or just use the previous full model run to initialize...
            results, lof = run_param_est(new_template, nfe, ncp, sigmas)
            initial_solutions.append((filt, lof))

            count += 1

        count = 0
        for x in initial_solutions:
            print(f'When wavelengths of less than {x[0]:0.3f} correlation are removed')
            print(f'The lack of fit is {x[1]:0.6f} %')

    # =============================================================================
    # --------------------------- DIAGNOSTIC TOOLS ------------------------
    # =============================================================================

    def lack_of_fit(self):
        """ Runs basic post-processing lack of fit analysis

        :return float lack of fit: percentage lack of fit

        """
        # Unweighted SSE - Difference between the calculated and measured D
        if self._spectra_given:
        
            sum_e = 0
            sum_d = 0
            C = np.zeros((len(self.model.times_spectral), self.n_val))
            S = np.zeros((len(self.model.meas_lambdas), self.n_val))
            
            for c_count, c in enumerate(self.comps['unknown_absorbance']):
                for t_count, t in enumerate(self.model.times_spectral):
                    C[t_count, c_count] = getattr(self.model, 'C')[t, c].value
    
            for c_count, c in enumerate(self.comps['unknown_absorbance']):
                for l_count, l in enumerate(self.model.meas_lambdas):
                    S[l_count, c_count] = self.model.S[l, c].value
                 
            D_model = C.dot(S.T)
            
            for t_count, t in enumerate(self.model.times_spectral):
                for l_count, l in enumerate(self.model.meas_lambdas):
                    sum_e += (D_model[t_count, l_count] - self.model.D[t, l]) ** 2
                    sum_d += (self.model.D[t, l]) ** 2
      
            lof = np.sqrt(sum_e/sum_d)*100
            
        return lof

    def wavelength_correlation(self):
        """ determines the degree of correlation between the individual wavelengths and
        the and the concentrations.


        :return dict: dictionary of correlations with wavelength

        """
        nt = len(self.model.allmeas_times)

        cov_d_l = dict()
        for c in self.comps['unknown_absorbance']:
            for l in self.model.meas_lambdas:
                mean_d = (sum(self.model.D[t, l] for t in self.model.times_spectral) / nt)
                mean_c = (sum(self.model.C[t, c].value for t in self.model.times_spectral) / nt)
                cov_d_l[l, c] = 0
                for t in self.model.times_spectral:
                    cov_d_l[l, c] += (self.model.D[t, l] - mean_d) * (self.model.C[t, c].value - mean_c)

                cov_d_l[l, c] = cov_d_l[l, c] / (nt - 1)

        # calculating the standard devs for dl and ck over time
        s_dl = dict()

        for l in self.model.meas_lambdas:
            s_dl[l] = 0
            mean_d = (sum(self.model.D[t, l] for t in self.model.times_spectral) / nt)
            error = 0
            for t in self.model.times_spectral:
                error += (self.model.D[t, l] - mean_d) ** 2
            s_dl[l] = (error / (nt - 1)) ** 0.5

        s_ck = dict()

        for c in self.comps['unknown_absorbance']:
            s_ck[c] = 0
            mean_c = (sum(self.model.C[t, c].value for t in self.model.times_spectral) / nt)
            error = 0
            for t in self.model.times_spectral:
                error += (self.model.C[t, c].value - mean_c) ** 2
            s_ck[c] = (error / (nt - 1)) ** 0.5

        cor_lc = dict()

        for c in self.comps['unknown_absorbance']:
            for l in self.model.meas_lambdas:
                cor_lc[l, c] = cov_d_l[l, c] / (s_ck[c] * s_dl[l])

        cor_l = dict()
        for l in self.model.meas_lambdas:
            cor_l[l] = max(cor_lc[l, c] for c in self.comps['unknown_absorbance'])

        return cor_l

    def lack_of_fit_huplc(self):
        """ Runs basic post-processing lack of fit analysis

        :return float lack of fit: percentage lack of fit

        """
        nc = len(self.model.huplc_absorbing)
        sum_e = 0
        sum_d = 0
        D_model = np.zeros((len(self.model.huplcmeas_times), nc))
        
        for t_count, t in enumerate(self.model.huplcmeas_times):
            for l_count, l in enumerate(self._list_huplcabs):
                
                if hasattr(self.model, 'solidvol'):
                    D_model[t_count, l_count] = (self.model.Z[t, l].value + self.model.solidvol[t, l].value) / (
                        sum(self.model.Z[t, j].value + self.model.solidvol[t, j].value for j in self._list_huplcabs))
                else:
                    D_model[t_count, l_count] = (self.model.Z[t, l].value) / (
                        sum(self.model.Z[t, j].value for j in self._list_huplcabs))

                sum_e += (D_model[t_count, l_count] - self.model.Dhat[t, l].value) ** 2
                sum_d += (self.model.Dhat[t, l].value) ** 2

        lof = np.sqrt((sum_e/sum_d))*100

        print(f"The lack of fit for the huplc data is {lof:0.2f}%")
        return lof
    
    def g_handling_status_messages(self):
        """Determines the unwanted contribution type and informs the user of this

        :return: None

        """
        if self.G_contribution == 'unwanted_G':
            print("\nType of unwanted contributions not set, so assumed that it is time-variant.\n")
            self.G_contribution = 'time_variant_G'
        elif self.G_contribution == 'time_variant_G':
            print("\nTime-variant unwanted contribution is involved.\n")
        elif self.G_contribution == 'time_invariant_G_decompose':
            print("\nTime-invariant unwanted contribution is involved and G can be decomposed.\n")
        elif self.G_contribution == 'time_invariant_G_no_decompose':
            print("\nTime-invariant unwanted contribution is involved but G cannot be decomposed.\n")
        return None
     
    def decompose_G_test(self, St, Z_in):
        """Check whether or not G can be decomposed

        :param dict St: Reaction coefficient matrix
        :param dict Z_in: Dosing points

        :return: None

        """
        if St == dict() and Z_in == dict():
            raise RuntimeError('Because time-invariant unwanted contribution is chosen, please provide information of St or Z_in to build omega matrix.')
        
        omega_list = [St[i] for i in St.keys()]
        omega_list += [Z_in[i] for i in Z_in.keys()]
        omega_sub = np.array(omega_list)
        rank = np.linalg.matrix_rank(omega_sub)
        cols = omega_sub.shape[1]
        rko = cols - rank
        
        if rko > 0:
            self.time_invariant_G_decompose = True
            self.G_contribution = 'time_invariant_G_decompose'
        else:
            self.time_invariant_G_no_decompose = True
            self.G_contribution = 'time_invariant_G_no_decompose'
        return None

def wavelength_subset_selection(correlations=None, n=None):
    """ identifies the subset of wavelengths that needs to be chosen, based
    on the minimum correlation value set by the user (or from the automated
    lack of fit minimization procedure)

    :param dict correlations: dictionary obtained from the wavelength_correlation  function, containing every
       wavelength from the original set and their correlations to the concentration profile.
    :param int n: a value between 0 - 1 that slects the minimum amount correlation between the wavelength and the
       concentrations.

    :return: A dictionary of correlations with wavelength
    :rtype: dict

    """
    if not isinstance(correlations, dict):
        raise RuntimeError("correlations must be of type dict()! Use wavelength_correlation function first!")

    # should check whether the dictionary contains all wavelengths or not
    if not isinstance(n, float):
        raise RuntimeError("n must be of type int!")
    elif n > 1 or n < 0:
        raise RuntimeError("n must be a number between 0 and 1")

    subset_dict = dict()
    for l in correlations.keys():
        if correlations[l] >= n:
            subset_dict[l] = correlations[l]
    return subset_dict


# =============================================================================
# ----------- PARAMETER ESTIMATION WITH WAVELENGTH SELECTION ------------------
# =============================================================================

def construct_model_from_reduced_set(builder_clone, end_time, D):
    """ constructs the new pyomo model based on the selected wavelengths.

    :param TemplateBuilder builder_clone: Template builder class of complete model without the data added yet
    :param float end_time: The end time for the data and simulation
    :param pandas.DataFrame D: The new, reduced dataset with only the selected wavelengths.

    :return TemplateBuilder opt_mode: new Pyomo model from TemplateBuilder, ready for parameter estimation

    """
    import pandas as pd
    from kipet.model_tools.template_builder import TemplateBuilder
    from kipet.model_components.spectral_handler import SpectralData
    
    
    if not isinstance(builder_clone, TemplateBuilder):
        raise RuntimeError('builder_clone needs to be of type TemplateBuilder')

    if not isinstance(D, pd.DataFrame):
        raise RuntimeError('Spectral data format not supported. Try pandas.DataFrame')

    builder_clone._spectral_data = D
    
    # spectral_data = SpectralData('D_new', data=D)
    # builder_clone.input_data(None, spectral_data)
    
    opt_model = builder_clone.create_pyomo_model(0.0, end_time, estimator='p_estimator')

    return opt_model


def run_param_est(opt_model, nfe, ncp, sigmas, solver='ipopt'):
    """ Runs the parameter estimator for the selected subset

    :param ConcreteModel opt_model: The model that we wish to run the
    :param int nfe: The number of finite elements
    :param int ncp: The number of collocation points
    :param dict sigmas: A dictionary containing the variances, as used in the ParameterEstimator class

    :return results_pyomo: Parameter Estimation results
    :return lof: lack of fit results

    """
    p_estimator = ParameterEstimator(opt_model)
    p_estimator.apply_discretization('dae.collocation', nfe=nfe, ncp=ncp, scheme='LAGRANGE-RADAU')
    options = dict()

    

    # These may not always solve, so we need to come up with a decent initialization strategy here
    results_pyomo = p_estimator.run_opt('ipopt',
                                        tee=False,
                                        solver_opts=options,
                                        variances=sigmas)
    # else:
    #     results_pyomo = p_estimator.run_opt(solver,
    #                                         tee=False,
    #                                         solver_opts=options,
    #                                         variances=sigmas,
    #                                         covariance=True)
    
    lof = p_estimator.lack_of_fit()

    return results_pyomo, lof



"""
Plotting class for KIPET
"""
# Standard library imports
from pathlib import Path
import os
import sys

# Third party imports 
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio

from kipet.model_tools.pyomo_model_tools import convert

pio.templates.default = "plotly_white"

# Kipet library imports
"""
Constants used for plotting
"""
# Default Matlab colors
colors_rgb = [(0,    0.4470,    0.7410),
          (0.8500,    0.3250,    0.0980),
          (0.9290,    0.6940,    0.1250),
          (0.4940,    0.1840,    0.5560),
          (0.4660,    0.6740,    0.1880),
          (0.3010,    0.7450,    0.9330),
          (0.6350,    0.0780,    0.1840)
          ]

# Convert to rgb format used in plotly
colors = ['rgb(' + ','.join([str(int(255*c)) for c in color]) + ')' for color in colors_rgb]

colors += ['#4285F4', '#DB4437', '#F4B400', '#0F9D58',
                     '#185ABC', '#B31412', '#EA8600', '#137333',
                     '#d2e3fc', '#ceead6']

# Use for making SVGs

# plot_options = {
#     'label_font': dict(
#             size=32,
#         ),
#     'title_font': dict(
#             size=32,
#         ),
#     'tick_font': dict(
#             size=24,
#         ),
#     }

plot_options = {
    'label_font': dict(
            size=14,
        ),
    'title_font': dict(
            size=18,
        ),
    'tick_font': dict(
            size=14,
        ),
    }


class PlotObject:

    """This will hold the relevant information needed to make a plot in KIPET

    This object is created in ReactionModel and accessed using the plot method therein.

    :param ReactionModel reaction_model: A ReactionModel instance
    :param bool jupyter: Indicates if the user is using a Jupyter notebook
    :param str filename: Optional file name for the plot
    """
    def __init__(self, reaction_model=None, jupyter=False, filename=None, show=False):
        """Initialization of the PlotObject instance

        :param ReactionModel reaction_model: A ReactionModel instance
        :param bool jupyter: Indicates if the user is using a Jupyter notebook
        :param str filename: Optional file name for the plot

        """
        self.reaction_model = reaction_model
        self.name = reaction_model.name
        self.color_num = 0
        self.filename = filename
        self.jupyter = jupyter
        self.show = show
        
        self.folder_name = self.reaction_model.timestamp
        if self.filename is not None:
            self.folder_name = self.filename
            
        if self.reaction_model.models['p_model']:
            self.results = self.reaction_model.results
        else:
            self.results = self.reaction_model.results_dict['simulator']
            

    @staticmethod
    def _make_line_trace(fig, x, y, name, color):
        """Convenience method for making traces in place.

        :param go.Figure fig: A figure object
        :param list x: X-axis values for plot
        :param list y: Y-axis values for plot
        :param str name: Name of the data
        :param int color: Index for the color

        :return: None
        
        """
        line = dict(color=colors[color], width=2)
        fig.add_trace(
            go.Scatter(x=x,
                       y=y,
                       name=name,
                       line=line,
               )
            )
        return None
    
    def _make_marker_trace(self, fig, x, y, name, color, marker_options):
        """Convenience method for making marker traces in place.

        :param go.Figure fig: A figure object
        :param list x: X-axis values for plot
        :param list y: Y-axis values for plot
        :param str name: Name of the data
        :param int color: Index for the color
        :param dict marker_options: Options for the marker characteristics

        :return: None
        
        """
        fig.add_trace(
            go.Scatter(x=x,
                       y=y,
                       name=name,
                       mode='markers',
                       marker={**marker_options, 'color':colors[self.color_num]}),
                    )
        return None

    def _fig_finishing(self, fig, pred, plot_name='Z', use_index=True, exp=None):
        """Finish the plots before showing.

        This method creates the plots and opens the browser to show them. It also saves the plots as SVGs in the
        same charts directory as the HTML versions.

        :param go.Figure fig: The figure object
        :param pandas.DataFrame pred: The predicted data from the model
        :param str plot_name: The name of the plot (based on data)

        :return: None

        """
        if use_index:
            x_data = [t for t in pred.index]
        else:
            if isinstance(pred, np.ndarray):
                 x_data = (np.min(pred), np.max(pred))
                 y_data = (np.min(exp), np.max(exp)) 
            else:
                x_data = (np.min(pred.values), np.max(pred.values))
                y_data = (np.min(exp.values), np.max(exp.values)) 
            y_axis_mod = 0.025*(float(y_data[-1]) - float(y_data[0]))
            fig.update_yaxes(range=[float(y_data[0])-y_axis_mod, float(y_data[-1])+y_axis_mod])

        x_axis_mod = 0.025*(float(x_data[-1]) - float(x_data[0]))
        fig.update_xaxes(range=[float(x_data[0])-x_axis_mod, float(x_data[-1])+x_axis_mod])
        
        fig.update_xaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e', titlefont=plot_options['label_font'], tickfont=plot_options['tick_font'])
        fig.update_yaxes(zeroline=True, zerolinewidth=2, zerolinecolor='#4e4e4e', titlefont=plot_options['label_font'], tickfont=plot_options['tick_font'])
        
        fig.update_layout(title_font=plot_options['title_font'])
        fig.update_layout(legend_font=plot_options['label_font'])
        
        filename = f'{plot_name}.html'
        # Change the folder directory
            
        stem = self.reaction_model.file.stem
        
        if self.jupyter:
            chart_dir = Path.cwd().joinpath('results', f'{stem}-{self.folder_name}' , 'charts', f'{self.name}')
            plot_method = pio.show
            fig.update_layout(         
                autosize=False,
                width=1200,
                height=800,
                margin=dict(
                    l=50,
                    r=50,
                    b=50,
                    t=50,
                    pad=4
                    ),
                )
        else:
        
            calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
            chart_dir = Path(calling_file_name).joinpath('results', f'{stem}-{self.folder_name}', 'charts', f'{self.name}')
            plot_method = pio.write_html
        
        chart_dir.mkdir(parents=True, exist_ok=True)
        filename = chart_dir.joinpath(filename)
        #if not self.jupyter:
        #   print(f'Plot saved as: {filename}')
           
        self.save_static_image = True
        if self.save_static_image:
            fig.write_image(f'{filename}.svg', width=1400, height=900)

        plot_method(fig,
                    file=filename.as_posix(),
                    auto_open=self.show,
                    include_plotlyjs='cdn')
    
        return filename

    def _state_plot(self, fig, var, pred, exp, use_spectral_format=False):
        """Generic method to plot state profiles

        :param go.Figure fig: The figure object
        :param str var: The variable to plot
        :param pandas.DataFrame pred: The predicted data from the model
        :param pandas.DataFrame exp: The experimental data
        :param bool use_spectral_format: For absorbance profiles True, otherwise False

        :return: None

        """
        self._make_line_trace(fig=fig,
                              x=pred.index,
                              y=pred[var],
                              name=var,
                              color=self.color_num,
                              )
        marker_options = {'size': 8,
                          'opacity': 0.5,
                          }
        label = 'exp.'   
        if use_spectral_format:
            marker_options = {'size': 8,
                              'opacity': 0.5,
                             }
            label = 'spectral'         
        if exp is not None and var in exp.columns:
            self._make_marker_trace(fig=fig,
                                    x=exp.index,
                                    y=exp[var],
                                    name=f'{var} ({label})',
                                    color=self.color_num,
                                    marker_options=marker_options,
                                    )
        return None


    def _plot_input_D_data(self):
        """Plot all input data concentration profiles
        
        """
        data = self.reaction_model.spectra.data

        fig = go.Figure()
        fig.add_trace(go.Surface(x=data.columns,
                                 y=data.index,
                                 z=data.values,
                                 ))

        fig.update_layout(scene=dict(
            xaxis_title='Wavelength',
            yaxis_title='Time',
            zaxis_title='Absorbance'),
            margin=dict(r=100, b=50, l=100, t=50),
            #title_text=f'{self.name}: Spectral Data',
            #title_font_size=plot_options['title_font']['size'],
            )

        calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
        chart_dir = Path(calling_file_name).joinpath('results', f'{self.reaction_model.file.stem}-{self.folder_name}', 'charts', f'{self.name}')
        plot_method = pio.write_html
        
        chart_dir.mkdir(parents=True, exist_ok=True)
        filename = chart_dir.joinpath('spectral_data.html')
        if not self.jupyter:
           print(f'Plot saved as: {filename}')
           
        self.save_static_image = True
        #if self.save_static_image:
        #    fig.write_image(f'{filename.as_posix()[:-5]}.svg', width=1400, height=900)

        plot_method(fig, file=filename.as_posix(), auto_open=False)
        
        return filename


    def _plot_all_Z(self):
        """Plot all concentration profiles
        
        """
        fig = go.Figure()
        use_spectral_format = False
        pred = getattr(self.results, 'Z')
        if hasattr(self.results, 'Cm'):
            exp = getattr(self.results, 'Cm')
        elif hasattr(self.results, 'C'):
            if self.reaction_model.models['_s_model'] and not self.reaction_model.models['v_model'] and not self.reaction_model.models['p_model']:
                exp = None
            else:
                exp = getattr(self.results, 'C')
                use_spectral_format = True
        else:
            exp = None
        for i, col in enumerate(pred.columns):
            self._state_plot(fig, col, pred, exp, use_spectral_format=use_spectral_format)
            self.color_num += 1
        self.color_num = 0
        var_data = self.reaction_model.components[col]
        state = f'{self.reaction_model.components[col].state}'.capitalize()
        title = f'Model: {self.reaction_model.name} | Concentration Profiles'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'

        state_units = self._get_proper_unit_str(var_data)
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name='all-concentration-profiles')

        return None

    def _plot_Z(self, var):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        use_spectral_format = False
        pred = getattr(self.results, 'Z')
        if hasattr(self.results, 'Cm'):
            exp = getattr(self.results, 'Cm')
            self._state_plot(fig, var, pred, exp)
        elif hasattr(self.results, 'C'):
            exp = getattr(self.results, 'C')
            use_spectral_format = True
            self._state_plot(fig, var, pred, exp, use_spectral_format=use_spectral_format)
        else:
            exp = None
        var_data = self.reaction_model.components[var]
        state = f'{self.reaction_model.components[var].state}'.capitalize()
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        state_units = self._get_proper_unit_str(var_data)
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-concentration-profile')
        
    def _plot_Y(self, var, extra=None):
        """Plot state profiles

        :param str var: algebraic variable

        :return: None
        """
        fig = go.Figure()
        pred = self.results.Y
        if hasattr(self.results, 'UD'):
            exp = self.results.UD
        else:
            exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.reaction_model.algebraics[var]
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        state_units = self._get_proper_unit_str(var_data, check_expr=True)
        fig.update_layout(
            title=title,
            xaxis_title=f'{time_scale}',
            yaxis_title=f'[{state_units}]',
            )
        self._fig_finishing(fig, pred, plot_name=f'{var}-profile')

    def _plot_X(self, var):
        """Plot state profiles
        
        :param str var: state variable

        :return: None

        """
        fig = go.Figure()
        pred = getattr(self.results, 'X')
        if hasattr(self.results, 'U'):
            exp = getattr(self.results, 'U')
        else:
            exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.reaction_model.states[var]
        description = f'| Description: {var_data.description}' if var_data.description is not None else ''
        title = f'Model: {self.reaction_model.name} | Variable: {var_data.name} {description}'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        state = f'{var_data.description}'.capitalize() if var_data.description is not None else 'State' 
        state_units = self._get_proper_unit_str(var_data)
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-state-profile')
        
    def _plot_all_S(self):
        """Plot all S profiles

        :return: None

        """
        fig = go.Figure()
        pred = getattr(self.results, 'S')
        exp = None
        for i, col in enumerate(pred.columns):
            self._state_plot(fig, col, pred, exp)
            self.color_num += 1
        self.color_num = 0
        description = f'Description: Single species absorbance profiles'
        title = f'Model: {self.reaction_model.name} | {description}'
        time_scale = f'Wavelength [centimeter]'
        state = 'Absorbance'
        state_units = 'liter / mol / centimeter'
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name='absorbance-spectra-all')

    def _plot_S(self, var, orig=False):
        """Plot individual S profile

        :param str var: component name

        :return: None

        """
        fig = go.Figure()
        if not orig:
            pred = getattr(self.results, 'S')
        else:
            pred = self.reaction_model.components[var].S
        exp = None
        self._state_plot(fig, var, pred, exp)
        var_data = self.results.S[var]
        description = f'| Description: Single species absorbance profiles'
        title = f'Model: {self.reaction_model.name} | Variable: {var} {description}'
        time_scale = f"Wavelength [centimeter]"
        state = 'Absorbance'
        state_units = 'liter / mol / centimeter'
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title=f'{state} [{state_units}]',
                )
        self._fig_finishing(fig, pred, plot_name=f'{var}-absorbance-spectra')
        
    def _plot_step(self, var):
        """Plot the step function

        :param str var: The step variable to plot

        :return: None

        """
        fig = go.Figure()
        pred = self.results.step
        self._state_plot(fig, var, pred, None)
        title = f'Model: {self.reaction_model.name} | Variable: {var} | Step Function'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        fig.update_layout(
            title=title,
            xaxis_title=f'{time_scale}',
            yaxis_title=f'[ - ]',
            )
        self._fig_finishing(fig, pred, plot_name=f'{var}-step-profile')

        return None


    def _residual_plot(self, fig, var, pred, exp, use_spectral_format=False):
        """Generic method to plot state profiles

        :param go.Figure fig: The figure object
        :param str var: The variable to plot
        :param pandas.DataFrame pred: The predicted data from the model
        :param pandas.DataFrame exp: The experimental data
        :param bool use_spectral_format: For absorbance profiles True, otherwise False

        :return: None

        """
        marker_options = {'size': 10,
                          'opacity': 0.8,
                          }
        if use_spectral_format:
            marker_options = {'size': 6,
                              'opacity': 0.4,
                          }    
        
        label = 'res.'   
        residuals = exp - pred
        
        self._make_marker_trace(fig=fig,
                                x=residuals.index,
                                y=residuals,
                                name=f'{var} ({label})',
                                color=self.color_num,
                                marker_options=marker_options,
                                )
        return None

    def _plot_Z_residuals(self):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        use_spectral_format = False
        pred = getattr(self.results, 'Z')
        exp = getattr(self.results, 'Cm')
        
        for i, col in enumerate(exp.columns):
            self._residual_plot(fig, col, pred[col], exp[col])
            self.color_num += 1
        self.color_num = 0
            
        title = f'Model: {self.reaction_model.name} | Concentration Residuals'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'

        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title='Residuals',
                )
        
        filename = self._fig_finishing(fig, pred, plot_name=f'concentration-residuals')
        return filename
    
    def _plot_D_residuals(self):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        use_spectral_format = False
        
        exp = convert(self.reaction_model.p_model.D)
        C = getattr(self.results, 'C')
        S = getattr(self.results, 'S')
        C = C.loc[:, S.columns]
        pred = C.dot(S.T)
        
        for i, col in enumerate(exp.columns):
            self._residual_plot(fig, col, pred[col], exp[col], use_spectral_format=True)
            #self.color_num += 1
        self.color_num = 0
            
        title = f'Model: {self.reaction_model.name} | Spectral Residuals'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'

        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title='Residuals',
                showlegend=False,
                )
        
        filename = self._fig_finishing(fig, pred, plot_name=f'spectral-residuals')
        return filename
        
        
    def _parity_plot(self, fig, var, pred, exp, use_spectral_format=False):
        """Generic method to plot state profiles

        :param go.Figure fig: The figure object
        :param str var: The variable to plot
        :param pandas.DataFrame pred: The predicted data from the model
        :param pandas.DataFrame exp: The experimental data
        :param bool use_spectral_format: For absorbance profiles True, otherwise False

        :return: None

        """
        marker_options = {'size': 8,
                          'opacity': 0.6,
                          }
        
        self._make_marker_trace(fig=fig,
                                x=pred,
                                y=exp,
                                name=f'{var}',
                                color=self.color_num,
                                marker_options=marker_options,
                                )
        
        return None
        
    def _plot_Z_parity(self):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        pred_raw = getattr(self.results, 'Z')
        exp = getattr(self.results, 'Cm')
        pred = pred_raw.loc[exp.index]
            
        line = dict(color='gray', width=2, dash='dash')
        fig.add_trace(
            go.Scatter(x=[0, np.max(pred.values)],
                       y=[0, np.max(pred.values)],
                       line=line,
               )
            )
            
        for i, col in enumerate(pred.columns):
            if col not in exp.columns:
                self.color_num += 1
                continue
            self._parity_plot(fig, col, pred[col], exp[col])
            self.color_num += 1
        self.color_num = 0

        title = f'Model: {self.reaction_model.name} | Concentration Parity'

        fig.update_layout(
                title=title,
                xaxis_title='Model Prediction',
                yaxis_title='Measured',
                autosize=False,
                width=550,
                height=550,
                )
        filename = self._fig_finishing(fig, pred, plot_name='concentration-parity', use_index=False, exp=exp)
        return filename
    
    def _plot_D_parity(self):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        use_spectral_format = False
       
        exp = convert(self.reaction_model.p_model.D)
        
        C = getattr(self.results, 'C')
        S = getattr(self.results, 'S')
        C = C.loc[:, S.columns]
        pred = C.dot(S.T)
        
        exp = exp.values.flatten()
        pred = pred.values.flatten()
        
        line = dict(color='gray', width=2, dash='dash')
        fig.add_trace(
            go.Scatter(x=[0, np.max(pred)],
                       y=[0, np.max(pred)],
                       line=line,
               )
            )
            
        self.color_num = 0
        self._parity_plot(fig, 'D', pred, exp)            
        
        title = f'Model: {self.reaction_model.name} | Spectral Parity'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        
        fig.update_layout(
                title=title,
                xaxis_title=f'Model Prediction',
                yaxis_title='Measured',
                autosize=False,
                width=550,
                height=550,
                )
        filename = self._fig_finishing(fig, pred, plot_name=f'spectral-parity', use_index=False, exp=exp)
        return filename
        
    def _plot_X_residuals(self, var):
        """Plot state profiles

        :param str var: concentration variable
        
        """
        fig = go.Figure()
        
        pred = getattr(self.results, 'X')
        if hasattr(self.results, 'U'):
            exp = getattr(self.results, 'U')
        else:
            exp = None
        
        self._residual_plot(fig, var, pred, exp)
        title = f'Model: {self.reaction_model.name} | State Residuals'
        time_scale = f'Time [{self.reaction_model.unit_base.time}]'
        fig.update_layout(
                title=title,
                xaxis_title=f'{time_scale}',
                yaxis_title='Residuals',
                )
        
        filename = self._fig_finishing(fig, pred, plot_name=f'{var}-state-residuals')
        return filename

    def _get_proper_unit_str(self, var_data, check_expr=False):
        """Gets the proper units for the charts labels

        :param str var_data: The variable data object (ModelElement)
        :param bool check_expr: Optional checking of expression units

        :return: The units
        :rtype: str

        """
        try:
            str_units = var_data.units.u
        except:
            str_units = var_data.units
            
        if check_expr:
            if var_data.name in self.reaction_model.alg_obj.exprs:
                str_units = str(self.reaction_model.alg_obj.exprs[var_data.name].units)
            else:
                if var_data.name in self.reaction_model.algebraics.names:
                    str_units = str(self.reaction_model.algebraics[var_data.name].units)
            
        return str_units



"""
Unit Base - Object for holding units and the unit registry
"""
import pint


class UnitBase:

    """Class to hold universal units and the unit registry

    :param pint.UnitRegistry ur: default unit registry
    :param str time: time base
    :param str volume: volume base
    :param str concentration: concentration base
    """

    def __init__(self):
        """Initialize the UnitBase for the ReactionModel

        :param pint.UnitRegistry ur: default unit registry
        :param str time: time base
        :param str volume: volume base
        :param str concentration: concentration base

        """
        self.ur = pint.UnitRegistry()
        self.time = 'min'
        self.volume = 'L'
        self.concentration = 'M'


"""
Model diagnostics
"""
import numpy as np

from kipet.model_tools.pyomo_model_tools import convert


def diagnostic_terms(exp, pred, num_params):
    """Calculates various model fit values using the experimental and
    predicted values
    
    :param pd.DataFrame exp: The experimental data
    :param pd.DataFrame pred: The predicted data from the model
    :param int num_params: The number of model parameters
    
    :return: Dict of various model fit values
    :rtype: dict
    
    """
    output = {}
    e = pred - exp
    n = exp.size
    
    # Sum of squared errors (also SSres)
    SSQ = (e ** 2).sum().sum()
    output['SSQ'] = (SSQ, 'Sum of squared errors')
    
    # Sum of absolute errors
    SAE = abs(e).sum().sum()
    output['SAE'] = (SAE, 'Sum of absolute errors')
    
    # Mean of SAE
    MAE = SAE/n
    output['MAE'] = (MAE, 'Mean absolute error')

    # Mean squared error
    MSE = SSQ/n
    output['MSE'] = (MSE, 'Mean squared error')

    # Root mean squared error
    RMSE = np.sqrt(MSE)
    output['RMSE'] = (RMSE, 'Root mean square error')
    
    # Relative absolute error
    y_bar = exp.sum().sum()/n
    y_bar_sum = abs(exp - y_bar).sum().sum()
    RAE = SAE/y_bar_sum
    output['RAE'] = (RAE, 'Relative absolute error')

    # Relative squared error
    SStot = ((exp - y_bar)**2).sum().sum()
    RSE = np.sqrt(SSQ/SStot)
    output['RSE'] = (RSE, 'Relative squared error')
    
    # R2
    R2 = 1 - SSQ/SStot
    output['R-sqaured'] = (R2, 'R squared')
    
    # Adj-R2
    R2_adj = 1 - (1 - R2)*(n - 1)/(n - num_params - 1)
    output['R-squared (adj)'] = (R2_adj, 'Adjusted R sqaured')
    
    # Residual Standard Error
    S = np.sqrt(SSQ/(n - num_params))
    output['S'] = (S, 'Residual standard error')
    
    return output


def model_fit(parameter_estimator):
    """ Runs basic post-processing lack of fit analysis

    :param ParameterEstimator parameter_estimator: The parameter estimator object after solving

    :return: Various model fit values
    :rtype: dict
    
    """
    model = parameter_estimator.model
    num_params = len(parameter_estimator.param_names)
    
    if hasattr(model, 'C'):
    
        C = convert(model.C)
        S = convert(model.S)
        C_red = C.loc[:, S.columns]
        exp = convert(model.D)
        pred = C_red.dot(S.T)
        
    elif hasattr(model, 'Cm'):

        exp = convert(model.Cm)
        raw_pred = convert(model.Z)
        pred = raw_pred.loc[exp.index]
        
    output = diagnostic_terms(exp, pred, num_params)
        
    return output


"""
Report generation class
"""

# Standard library imports
from pathlib import Path
import os
import re 
import webbrowser

# Third party imports 
from jinja2 import Environment, FileSystemLoader
import pandas as pd
from pytexit import py2tex


class Report:
    
    """ This class contains the methods used to prepre model data for inclusion in the HTML report"""
    
    def __init__(self, model_list):
        
        self.reactions = model_list # ReactionModels

    @staticmethod
    def model_context(reaction_model):
        """Prepares a dictionary of high-level model attributes
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        
        :return model: Dictionary of attributes
        :rtype: dict
        
        """
        model = {}
        model['name'] = reaction_model.name
        model['n_comps'] = reaction_model.components
        model['n_states'] = reaction_model.states
        
        return model
    
    @staticmethod
    def component_context(reaction_model):
        """Prepares a dictionary of component model attributes
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        
        :return comps: Dictionary of attributes
        :rtype: dict
        
        """
        comps = []
        
        for comp in reaction_model.components:
            comp_data = {}
            comp_data['name'] = comp.name
            comp_data['units'] = dimensionless_check(comp.units)
            comp_data['value'] = comp.value
            comp_data['variance'] = comp.variance
            comp_data['known'] = 'Yes' if comp.known else 'No'
            comp_data['absorbing'] = 'Yes' if comp.absorbing else 'No'
            comp_data['description'] = 'Not provided' if comp.description is None else comp.description
            comps.append(comp_data)
            
        return comps
    
    @staticmethod
    def state_context(reaction_model):
        """Prepares a dictionary of state model attributes
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        
        :return comps: Dictionary of attributes
        :rtype: dict
        
        """
        comps = []
        
        for comp in reaction_model.states:
            comp_data = {}
            comp_data['name'] = comp.name
            comp_data['units'] = dimensionless_check(comp.units)
            comp_data['value'] = comp.value
            comp_data['variance'] = comp.variance
            comp_data['known'] = 'Yes' if comp.known else 'No'
            comp_data['description'] = 'Not provided' if comp.description is None else comp.description
            comps.append(comp_data)
            
        return comps
    
    @staticmethod
    def parameter_context(reaction_model):
        """Prepares a dictionary of parameter model attributes
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        
        :return params: Dictionary of attributes
        :rtype: dict
        
        """
        params = []

        if reaction_model.models['p_model']:
            results_obj = reaction_model.results
        else:
            results_obj = reaction_model.results_dict['simulator']

        for param in reaction_model.parameters:
            param_data = {}
            param_data['name'] = param.name
            param_data['units'] = dimensionless_check(param.units)
            param_data['initial'] = param.value
            param_data['value'] = results_obj.P[param.name]
            param_data['lb'] = param.bounds[0]
            param_data['ub'] = param.bounds[1]
            param_data['description'] = 'Not provided' if param.description is None else param.description
            params.append(param_data)
            
            if hasattr(reaction_model, 'p_model'):
                param_data['fixed'] = reaction_model.p_model.P[param.name].fixed
            
        if hasattr(results_obj, 'time_step_change'):
            for indx, param in pd.DataFrame(results_obj.time_step_change).iterrows():
               
                param_data = {}
                param_data['name'] = indx
                param_data['units'] = reaction_model.unit_base.time
                param_data['value'] = param.loc[0]
                param_data['initial'] = reaction_model._s_model.time_step_change[indx].value
                param_data['lb'] = reaction_model._s_model.time_step_change[indx].bounds[0]
                param_data['ub'] = reaction_model._s_model.time_step_change[indx].bounds[1]
                param_data['description'] = 'Binary variable'
                
                if hasattr(reaction_model, 'p_model'):
                    param_data['fixed'] = reaction_model.p_model.time_step_change[indx].fixed
    
                params.append(param_data)
                
        if hasattr(results_obj, 'Pinit'):
            for indx, param in pd.DataFrame(results_obj.Pinit).iterrows():
               
                param_data = {}
                param_data['name'] = f'{indx}'
                param_data['units'] = dimensionless_check(reaction_model.components[indx].units)
                param_data['value'] = param.loc[0]
                param_data['lb'] = reaction_model.components[indx].bounds[0]
                param_data['ub'] = reaction_model.components[indx].bounds[1]
                param_data['description'] = 'Initial value'
    
                params.append(param_data)
            
        return params

    @staticmethod
    def constant_context(reaction_model):
        """Prepares a dictionary of constant model attributes
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        
        :return params: Dictionary of attributes
        :rtype: dict
        
        """
        params = []
        
        for param in reaction_model.constants:
            param_data = {}
            param_data['name'] = param.name
            param_data['units'] = dimensionless_check(param.units)
            param_data['value'] = param.value
            param_data['description'] = 'Not provided' if param.description is None else param.description
            params.append(param_data)
            
        return params

    @staticmethod
    def remove(text):
        """Removes brackets from strings
        
        :param str text: Text to be cleaned
        
        :return: Cleaned text
        :rtype: str
        
        """
        return re.sub("[\[].*?[\]]", "", text)
    

    def readable_odes(self, reaction_model, kind):
        """Converts the ODEs to LaTeX
        
        :param ReactionModel reaction_model: A solved/simulated ReactionModel instance
        :param str kind: odes or daes
        
        :return odes_new: Dictionary of equations as latex strings
        :rtype: dict
        
        """
        numeric_const_pattern = '[-+]? (?: (?: \d* \. \d+ ) | (?: \d+ \.? ) )(?: [Ee] [+-]? \d+ ) ?'
        rx = re.compile(numeric_const_pattern, re.VERBOSE)
        
        odes = {k: v.expression for k, v in getattr(reaction_model, kind).items()}
        odes_new = {}
        
        for k, v in odes.items():
            
            if kind == 'odes':
                try:
                    odes_new[f'{k}'] = self.remove(v.to_string())
                except:
                    odes_new[f'{k}'] = str(v)
            else:
                try:
                    odes_new[f'{k}'] = self.remove(v.to_string())
                except:
                    odes_new[f'{k}'] = str(v)
                    
            # floats = rx.findall(odes_new[k])

            # text = odes_new[k]
            # #print(text)
            # for num in floats:
            #     find_it = num
            #     if int(float(num)) == float(num):
            #         continue
            #     elif len(num) < 5:
            #         continue
            #     else:
            #         repl_it = f'{float(find_it):.4f}'
            #         text = text.replace(find_it, repl_it)

            odes_new[k] = py2tex(odes_new[k], print_latex=False, print_formula=False)
             
        ode_list = []
        for k, v in odes_new.items():
            ode_list.append(f'{k} = {v}')
            
        return odes_new
    
        
    def generate_report(self):
        """Generates a full HTML report for the KIPET model.
        
        This provides a more useful interface for representing results than using individual charts.
        Also included are tables with useful diagnostics and other results.
        
        :return: None
        
        """
        from kipet import __version__ as version
        model_dict = {}
        all_names = [rm.name for rm in self.reactions]
    
        user_file = self.reactions[0].file
        user_stem = user_file.stem
        time = self.reactions[0].timestamp
        current_dir = Path(__file__).parent
        templates_dir = (current_dir / 'templates').resolve()
        style_file = (templates_dir / 'report.css').resolve()
        prism_style_file = (templates_dir / 'prism.css').resolve()
        prism_js_file = (templates_dir / 'prism.js').resolve()
        
        style_file_text = ''
        with open(style_file, 'r+') as f:
            style_file_text = f.read()
        
        prism_css = ''
        with open(prism_style_file, 'r+') as f:
            prism_css = f.read()
        
        prism_js = ''
        with open(prism_js_file, 'r+') as f:
            prism_js = f.read()
        
        source_code = ''
        with open(user_file, 'r+') as f:
            source_code = f.read()
        
        
        user_dir = Path(user_file).parent
        filename = (user_dir / 'results' / f'{user_stem}-{time}' / 'report.html').resolve()
        
        env = Environment( loader = FileSystemLoader(templates_dir) )
        template = env.get_template('index.html')

        
        problem_text = "This is an automatically generated report presenting the results found using KIPET."
             
        log_orig = (user_dir / f'log-{user_stem}-{time}.txt').resolve()
        log_file = (user_dir / 'results' / f'{user_stem}-{time}' /  'log.txt').resolve()
        
        if log_orig.is_file():
            log_orig.rename(log_file)
        
        with open(log_file, 'r') as f:
            log = f.read()
        
        for reaction_model in self.reactions:
            
            name = reaction_model.name
            model_dict[name] = {}
            model_dict[name]['time'] = reaction_model.timestamp
            charts_dir = (user_dir / 'results'/  f'{user_stem}-{time}' / 'charts' / f'{reaction_model.name}').resolve()
            
            # Concentration profiles
            model_dict[name]['chart_C_files'] = [x for x in charts_dir.glob('all-concentration-profiles.html') if x.is_file()]
            model_dict[name]['chart_S_files'] = [x for x in charts_dir.glob('absorbance-spectra-all.html') if x.is_file()]
            model_dict[name]['chart_U_files'] = sorted([x for x in charts_dir.glob('*state-profile.html') if x.is_file()])
            model_dict[name]['chart_Y_files'] = sorted([x for x in charts_dir.glob('*profile.html') if x.is_file() and x not in model_dict[name]['chart_U_files']])
            
            data_chart_files = None
            spectral_info = None
            spectra_file = None
            
            if reaction_model.spectra is not None:
                spectra_file = reaction_model.spectra.file if reaction_model.spectra.file is not None else 'Not provided or custom'
                data_chart_files = reaction_model._plot_object._plot_input_D_data()
                
                sd = reaction_model.spectra
                spectral_info = []
                for method in ['_sg', '_snv', '_msc', '_base', '_decreased_wavelengths', '_decreased_times']:
                    if getattr(sd, method) is not None:
                        spectral_info.append(getattr(sd, method))
                        
            model_dict[name]['data_chart_files'] = data_chart_files
            model_dict[name]['spectral_info'] = spectral_info
            model_dict[name]['spectra_file'] = spectra_file
                        
            abs_data = None
            S_data = []
            for comp in reaction_model.components:
                if comp.S is not None:
                    reaction_model._plot_object._plot_S(comp.name, orig=True)
                    S_data.append(comp)
                    abs_data = 1
                    
            model_dict[name]['abs_data'] = abs_data
                    
            chart_abs_files = None
            if len(S_data) > 0:
                chart_abs_files = sorted([x for x in charts_dir.glob('*absorbance-spectra.html') if x.is_file()])
            
            model_dict[name]['chart_abs_files'] = chart_abs_files
            model_dict[name]['comp_data'] = self.component_context(reaction_model)
            model_dict[name]['param_data'] = self.parameter_context(reaction_model)
            model_dict[name]['const_data'] = self.constant_context(reaction_model)
            model_dict[name]['state_data'] = self.state_context(reaction_model)
            model_dict[name]['bounds'] = reaction_model._builder._prof_bounds
            model_dict[name]['variances'] = reaction_model.variances


            if reaction_model.models['p_model']:
                results_obj = reaction_model.results
            else:
                results_obj = reaction_model.results_dict['simulator']

            model_dict[name]['confidence'] = results_obj.deviations()
            model_dict[name]['covariance'] = results_obj.parameter_covariance
            
            
            # Figuring out what was done
            final_estimator = 'simulation'
            if reaction_model.models['p_model']:
                final_estimator = 'parameter estimation'
            elif reaction_model.models['v_model']:
                final_estimator = 'variance estimation'
                
            model_dict[name]['final_estimator'] = final_estimator
            
            used_variance = False
            if reaction_model.models['v_model']:
                used_variance = reaction_model.models['v_model']
            
            opt_data = {
                'var_method': reaction_model.settings.variance_estimator.method,
                'variance_used': used_variance,
                'final_method': final_estimator,
                'is_simulation': final_estimator == 'simulation',
                'var_add': '',
                }
            
            model_dict[name]['opt_data'] = opt_data
            
            if reaction_model.settings.variance_estimator.fixed_device_variance is not None:
                model_dict[name]['opt_data']['var_add'] = '- using fixed device variance'
        
        
            if reaction_model.settings.variance_estimator.method == 'direct_sigmas':
                mod_direct_sigma_dict = {}
                for key, value in reaction_model.direct_sigma_dict.items():
                    mod_direct_sigma_dict[key] = {}
                    mod_direct_sigma_dict[key]['delta'] = reaction_model.direct_sigma_dict[key]['delta']
                    mod_direct_sigma_dict[key].update(**reaction_model.direct_sigma_dict[key]['simgas'])
                model_dict[name]['delta_results'] = mod_direct_sigma_dict
            else:
                model_dict[name]['delta_results'] = None
            
            
            # Settings - make a series of tables for the settings
            settings = reaction_model.settings.as_dicts
            if 'solver_opts' in settings['simulator'][1]:
                settings['simulator'][1].pop('solver_opts')
            if 'solver_opts' in settings['variance_estimator'][1]:
                settings['variance_estimator'][1].pop('solver_opts')
            if 'solver_opts' in settings['parameter_estimator'][1]:
                settings['parameter_estimator'][1].pop('solver_opts')
            
            model_dict[name]['settings'] = settings
            model_dict[name]['ode_data'] = self.readable_odes(reaction_model, 'odes')
            
            age_data = None
            if hasattr(reaction_model, 'algs') and isinstance(getattr(reaction_model, 'algs'), dict) and len(getattr(reaction_model, 'algs')) > 0:
                age_data = {}
                all_algs = self.readable_odes(reaction_model, 'algs')
                all_rxns = reaction_model.algebraics.get_match('is_reaction', True)
                
                for k, eq in all_algs.items():
                    if k in all_rxns:
                        age_data[k] = ('Yes', eq)
                    else:
                        age_data[k] = ('No', eq)
            
            model_dict[name]['age_data'] = age_data
            
            from kipet.model_tools.diagnostics import model_fit

            diagnostics = None
            if hasattr(reaction_model, 'p_estimator'):

                diags = model_fit(reaction_model.p_estimator)
                diagnostics = []
                if final_estimator == 'parameter estimation':
                    for k, v in diags.items():
                        stat = {}
                        stat['name'] = k
                        stat['value'] = v[0]
                        stat['description'] = v[1]
                        diagnostics.append(stat)
                    
            model_dict[name]['diagnostics'] = diagnostics
                
            res_chart_files = None
            par_chart_files = None
            if reaction_model._builder._concentration_data is not None:
                res_chart_files = reaction_model._plot_object._plot_Z_residuals()
                par_chart_files = reaction_model._plot_object._plot_Z_parity()
                
            if reaction_model._builder._spectral_data is not None:
                res_chart_files = reaction_model._plot_object._plot_D_residuals()
                par_chart_files = reaction_model._plot_object._plot_D_parity()
                
            model_dict[name]['res_chart'] = res_chart_files
            model_dict[name]['par_chart'] = par_chart_files
            
            feeds = None
            dosing_dict = reaction_model._dosing_points
            if dosing_dict is not None:
                feeds = [x.as_list for x in dosing_dict['Z']]
                
            model_dict[name]['feeds'] = feeds
            
            model_dict[name]['g_contrib'] = None
            if hasattr(reaction_model, '_G_contribution') and reaction_model._G_contribution is not None:
                model_dict[name]['g_contrib'] = reaction_model._G_contribution
            
            self.model_dict = model_dict
        
        with open(filename, 'w') as fh:
            fh.write(template.render(
                user_file = user_file,
                file_stem = user_stem,
                default = all_names[0],
                version = version,
                models = model_dict,
                filename = filename,
                description = problem_text,
                log_file = log_file,
                log = log,
                base_text = style_file_text,
                prism_text = prism_css,
                prism_js = prism_js,
                source_code = source_code,
                time = time,
            ))
        
        webbrowser.open('file://' + os.path.realpath(filename))
            
        return None


def dimensionless_check(units):
    """Converts dimensionless to a hyphen
    
    :param units: The model component units
    
    :return: The formatted unit string
    :rtype: str
    
    """
    if units.u == '':
        return '-'
    else:
        return units.u
    

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kipet',
 'kipet.calculation_tools',
 'kipet.estimability_tools',
 'kipet.estimator_tools',
 'kipet.general_settings',
 'kipet.input_output',
 'kipet.main_modules',
 'kipet.mixins',
 'kipet.model_components',
 'kipet.model_tools',
 'kipet.variance_methods',
 'kipet.visuals']

package_data = \
{'': ['*'],
 'kipet': ['.idea/.gitignore',
           '.idea/.gitignore',
           '.idea/.gitignore',
           '.idea/.gitignore',
           '.idea/.gitignore',
           '.idea/inspectionProfiles/*',
           '.idea/kipet.iml',
           '.idea/kipet.iml',
           '.idea/kipet.iml',
           '.idea/kipet.iml',
           '.idea/kipet.iml',
           '.idea/misc.xml',
           '.idea/misc.xml',
           '.idea/misc.xml',
           '.idea/misc.xml',
           '.idea/misc.xml',
           '.idea/modules.xml',
           '.idea/modules.xml',
           '.idea/modules.xml',
           '.idea/modules.xml',
           '.idea/modules.xml',
           '.idea/vcs.xml',
           '.idea/vcs.xml',
           '.idea/vcs.xml',
           '.idea/vcs.xml',
           '.idea/vcs.xml'],
 'kipet.visuals': ['templates/prism.css',
                   'templates/prism.css',
                   'templates/prism.css',
                   'templates/prism.js',
                   'templates/prism.js',
                   'templates/prism.js',
                   'templates/report.css',
                   'templates/report.css',
                   'templates/report.css']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'Pint>=0.16.1,<0.17.0',
 'PyYAML>=5.4.1,<6.0.0',
 'Pygments>=2.9.0,<3.0.0',
 'Pyomo>=5.7.3,<6.0.0',
 'Sphinx>=4.0.1,<5.0.0',
 'attrs>=20.3.0,<21.0.0',
 'kaleido==0.2.1',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.2,<2.0.0',
 'plotly>=4.14.3,<5.0.0',
 'pytexit>=0.3.4,<0.4.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'scipy>=1.6.0,<2.0.0',
 'six>=1.15.0,<2.0.0']

setup_kwargs = {
    'name': 'kipet',
    'version': '1.0.2',
    'description': 'An all-in-one tool for fitting kinetic models using spectral and other state data',
    'long_description': '# <img alt="KIPET" src="branding/kipetlogo_full.svg" height="60">\n\n[![](https://img.shields.io/github/license/salvadorgarciamunoz/kipet)](https://github.com/salvadorgarciamunoz/kipet/blob/master/LICENSE)\n[![](https://img.shields.io/github/last-commit/salvadorgarciamunoz/kipet)](https://github.com/salvadorgarciamunoz/kipet/)\n[![](https://img.shields.io/pypi/wheel/kipet)](https://pypi.org/manage/project/kipet/release/0.1.1/)\n\n\nKIPET is a Python package designed to simulate, and estimate parameters from \nchemical reaction systems through the use of maximum likelihood principles,\nlarge-scale nonlinear programming and discretization methods. \n\n- **Documentation:** - https://kipet.readthedocs.io\n- **Examples and Tutorials** - https://github.com/kwmcbride/kipet_examples\n- **Source code:** - https://github.com/salvadorgarciamunoz/kipet\n- **Bug reports:** - https://github.com/salvadorgarciamunoz/kipet/issues\n\nIt has the following functionality:\n\n - Simulate a reactive system described with DAEs\n - Solve the DAE system with collocation methods\n - Pre-process data\n - Estimate variances of noise from the model and measurements\n - Estimate kinetic parameters from spectra or concentration data across 1 or \n  multiple experiments with different conditions\n - Estimate confidence intervals of the estimated parameters\n - Able to estimate variances and parameters for problems where there is dosing / inputs into the system\n - Provide a set of tools for estimability analysis\n - Allows for wavelength selection of most informative wavelengths from a dataset\n - Visualize results\n\n\n<br>\n\n## Installation\n\nThere are many options for installing KIPET.\n\n### PyPi\n[![](https://img.shields.io/badge/Install%20with-pip-green)]()\n[![](https://img.shields.io/pypi/v/kipet.svg?style=flat)](https://pypi.org/pypi/kipet/)\n<br>\n\nA packaged version of KIPET can be installed using:\n\n    pip install kipet\n\nIf you run into errors when installing KIPET using pip (such as when installing into a clean virtual environment), try installing the following packages beforehand:\n\n    pip install Cython numpy six\n    pip install kipet\n\n### Anaconda \n[![](https://anaconda.org/kwmcbride/kipet/badges/installer/conda.svg)]()\n[![Anaconda-Server Badge](https://img.shields.io/conda/vn/kwmcbride/kipet)](https://anaconda.org/kwmcbride/kipet)\n[![](https://img.shields.io/conda/pn/kwmcbride/kipet?color=orange)]()\n\n\nIf you are using Anaconda, KIPET can be installed using:\n\n    conda install -c kwmcbride kipet\n\nThe anaconda packages have the benefit of including pynumero ready to go, which is needed for some of the methods included in KIPET. You will need to compile these on your own if you choose to install KIPET using a different method. See the [pynumero readme](https://github.com/Pyomo/pyomo/tree/master/pyomo/contrib/pynumero) for more information. Otherwise, you can also use [k_aug](https://github.com/dthierry/k_aug) for these methods as well. \n\n### Poetry\n\nYou may also install KIPET with poetry:\n\n    poetry add kipet\n\n\n### GitHub\n\nAdditionally, KIPET may be installed directly from the repository (for example, if using poetry, simply install the desired branch (#branch) in the following manner):\n\n    poetry add git+https://github.com/salvadorgarciamunoz/kipet#master\n\nNaturally you can simply clone or download the repository if you wish.\n\n    cd <installation directory>\n    git clone https://github.com/salvadorgarciamunoz/kipet.git\n    cd kipet\n    python setup.py install\n\n### Ipopt and k_aug\n\nTo use KIPET to its full potential, you should install Ipopt and k_aug. Ipopt is \na popular solver for non-linear programs and k_aug is a new method to calculate sensitivities\nfrom the KKT matrix. The latter is required if covariances are to be calculated.\n\nTODO: add links to the installation of each here.\n\nTo help ease the installation of these software tools, there are two scripts written for\nLinux OS (Debian) for installing Ipopt and k_aug. These are available here https://github.com/kwmcbride/Linux-Setup-Scripts\n\n### Examples and Tutorials\n\nAll of the example problems can be easily downloaded from the examples repository:\n\n    cd <example directory>\n    git clone https://github.com/kwmcbride/kipet_examples.git\n\n\nTo validate your installation, you can now run the test script included with the examples:\n\n    cd <example directory>/kipet_examples\n    python run_examples.py\n\n<br>\n\n## License\n\nGPL-3\n\n\n## Authors\n\n    - Kevin McBride - Carnegie Mellon University\n    - Kuan-Han Lin - Carnegie Mellon University\n    - Christina Schenk - Basque Center for Applied Mathematics\n    - Michael Short - University of Surrey\n    - Jose Santiago Rodriguez - Purdue University\n    - David M. Thierry - Carnegie Mellon University\n    - Salvador García-Muñoz - Eli Lilly\n    - Lorenz T. Biegler - Carnegie Mellon University\n\n## Please cite\n<br>\n\n - C. Schenk, M. Short, J.S. Rodriguez, D. Thierry, L.T. Biegler, S. García-Muñoz, W. Chen (2020)\nIntroducing KIPET: A novel open-source software package for kinetic parameter estimation from experimental datasets including spectra, Computers & Chemical Engineering, 134, 106716. https://doi.org/10.1016/j.compchemeng.2019.106716\n\n - M. Short, L.T. Biegler, S. García-Muñoz, W. Chen (2020)\nEstimating variances and kinetic parameters from spectra across multiple datasets using KIPET, Chemometrics and Intelligent Laboratory Systems, https://doi.org/10.1016/j.chemolab.2020.104012\n\n - M. Short, C. Schenk, D. Thierry, J.S. Rodriguez, L.T. Biegler, S. García-Muñoz (2019)\nKIPET–An Open-Source Kinetic Parameter Estimation Toolkit, Computer Aided Chemical Engineering, 47, 299-304.\n\n\n\n\n\n\n',
    'author': 'Kevin McBride',
    'author_email': 'kevin.w.mcbride.86@gmail.com',
    'maintainer': 'Kevin McBride',
    'maintainer_email': 'kevin.w.mcbride.86@gmail.com',
    'url': 'https://github.com/kwmcbride/kipet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
This module contains the code related to jumps - removes redundancy

Changed to not being a mixin to clean up inheritance
"""
from pyomo.core import Var, Param, Constraint, ConstraintList
from pyomo.dae import ContinuousSet

from kipet.model_tools.visitor_classes import ReplacementVisitor

    
def set_up_jumps(model, kwargs):
    """Takes the kwargs from the Estimator classes and initializes the dosing points (or any other type
    of sudden change in the states)

    :param ConcreteModel model: Pyomo model (from Variance or Parameter Estimator)
    :param dict var_dict: Dict of states
    :param dict jump_times: Dict of jump times
    :param array-like feed_times: List of feed times (could be a set)

    :return: None
    
    """
    var_dict = kwargs.pop("jump_states", None)
    jump_times = kwargs.pop("jump_times", None)
    feed_times = kwargs.pop("feed_times", None)
    
    if not isinstance(var_dict, dict):
        print("disc_jump_v_dict is of type {}".format(type(var_dict)))
        raise Exception  # wrong type
    if not isinstance(jump_times, dict):
        print("disc_jump_times is of type {}".format(type(jump_times)))
        raise Exception  # wrong type
    count = 0
    
    for i in jump_times.keys():
        for j in jump_times[i].items():
            count += 1
    if len(feed_times) > count:
        raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                    "There are more time points in feed_times than jump_times provided.")

    time_set = None
    for i in model.component_objects(ContinuousSet):
        time_set = i
        break
    if time_set is None:
        raise Exception('No continuous_set')

    time_set = time_set.name

    ttgt = getattr(model, time_set)
    ncp = ttgt.get_discretization_info()['ncp']
    fe_l = ttgt.get_finite_elements()
    fe_list = [fe_l[i + 1] - fe_l[i] for i in range(0, len(fe_l) - 1)]

    for fe in range(0, len(fe_list)):
        
        vs = ReplacementVisitor()
        kn = 0
        for ki in jump_times.keys():
            if not isinstance(ki, str):
                print("ki is not str")
            vtjumpkeydict = jump_times[ki]
            for l in vtjumpkeydict.keys():
                jump_time = vtjumpkeydict[l]
                jump_fe, jump_cp = fe_cp(ttgt, jump_time)
                if jump_time not in feed_times:
                    raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                                    "They do not match.\n"
                                    "Jump_time is not included in feed_times.")
                if fe == jump_fe + 1:
                    for v in var_dict.keys():
                        if not isinstance(v, str):
                            print("v is not str")
                        vkeydict = var_dict[v]
                        for k in vkeydict.keys():
                            if k == l:  # Match in between two components of dictionaries
                                var = getattr(model, v)
                                #dvar = getattr(model, "d" + v + "dt")
                                con_name = 'd' + v + 'dt_disc_eq'
                                con = getattr(model, con_name)

                                model.add_component(v + "_dummy_eq_" + str(kn), ConstraintList())
                                conlist = getattr(model, v + "_dummy_eq_" + str(kn))
                                varname = v + "_dummy_" + str(kn)
                                model.add_component(varname, Var([0]))  #: this is now indexed [0]
                                vdummy = getattr(model, varname)
                                vs.change_replacement(vdummy[0])   #: who is replacing.
                                jump_delta = vkeydict[k]
                                model.add_component(v + '_jumpdelta' + str(kn),
                                                         Param(initialize=jump_delta))
                                jump_param = getattr(model, v + '_jumpdelta' + str(kn))
                                if not isinstance(k, tuple):
                                    k = (k,)
                                exprjump = vdummy[0] - var[(jump_time,) + k] == jump_param  #: this cha
                                # exprjump = vdummy - var[(self.jump_time,) + k] == jump_param
                                model.add_component("jumpdelta_expr" + str(kn), Constraint(expr=exprjump))
                                for kcp in range(1, ncp + 1):
                                    curr_time = t_ij(ttgt, jump_fe + 1, kcp)
                                    if not isinstance(k, tuple):
                                        knew = (k,)
                                    else:
                                        knew = k
                                    idx = (curr_time,) + knew
                                    con[idx].deactivate()
                                    e = con[idx].expr
                                    suspect_var = e.args[0].args[1].args[0].args[0].args[1]  #: seems that
                                    vs.change_suspect(id(suspect_var))  #: who to replace
                                    e_new = vs.dfs_postorder_stack(e)  #: replace
                                    con[idx].set_value(e_new)
                                    conlist.add(con[idx].expr)
                kn = kn + 1

def t_ij(time_set, i, j):
    """Return the corresponding time(continuous set) based on the i-th finite element and j-th collocation point
    From the NMPC_MHE framework by @dthierry.

    :param ContinuousSet time_set: Parent Continuous set
    :param int i: finite element
    :param int j: collocation point

    :return: Corresponding index of the ContinuousSet
    :rtype: float

    """
    if i < time_set.get_discretization_info()['nfe']:
        h = time_set.get_finite_elements()[i + 1] - time_set.get_finite_elements()[i]  #: This would work even for 1 fe
    else:
        h = time_set.get_finite_elements()[i] - time_set.get_finite_elements()[i - 1]  #: This would work even for 1 fe
    tau = time_set.get_discretization_info()['tau_points']
    fe = time_set.get_finite_elements()[i]
    time = fe + tau[j] * h
    return round(time, 6)


def fe_cp(time_set, feedtime):
    """Return the corresponding fe and cp for a given time

    :param ContinuousSet time_set: Parent ContinuousSet
    :param float feedtime: Time of the step change
    
    :return fe: The finite element
    :return cp: The collocation point
    :rtype tuple:
             
    """
    fe_l = time_set.get_lower_element_boundary(feedtime)
    fe = None
    j = 0
    for i in time_set.get_finite_elements():
        if fe_l == i:
            fe = j
            break
        j += 1
    h = time_set.get_finite_elements()[1] - time_set.get_finite_elements()[0]
    tauh = [i * h for i in time_set.get_discretization_info()['tau_points']]
    j = 0  #: Watch out for LEGENDRE
    cp = None
    for i in tauh:
        if round(i + fe_l, 6) == feedtime:
            cp = j
            break
        j += 1
    return fe, cp



"""Multiple Experiment Estimator"""

# Standard library imports
import copy

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import (Block, ConcreteModel, Constraint, minimize, 
                           Objective, SolverFactory, Var)

# KIPET library imports
from kipet.model_components.objectives import (absorption_objective, comp_objective,
                                               conc_objective)
from kipet.estimator_tools.reduced_hessian_methods import define_free_parameters
from kipet.estimator_tools.results_object import ResultsObject
from kipet.mixins.parameter_estimator_mixins import PEMixins
from kipet.general_settings.variable_names import VariableNames


class MultipleExperimentsEstimator(PEMixins, object):
    """This class is for Estimation of Variances and parameters when we have multiple experimental datasets.
    This class relies heavily on the Pyomo block class as we put each experimental class into its own block.
    This blocks are first run individually in order to find good initializations and then they are linked and
    run together in a large optimization problem.

    :param list reaction_models: The full model TemplateBuilder problem needs to be fed into the
         MultipleExperimentsEstimator. This pyomo model will form the basis for all the optimization tasks

    """
    def __init__(self, reaction_models):
        
        self.reaction_models = reaction_models
        
        self.experiments = list(self.reaction_models.keys())
        self.variances = {name: model.variances for name, model in self.reaction_models.items()}
        self.global_params = None
        self.parameter_means = False
        self.rm_cov = None
        self.__var = VariableNames()

    
    def _display_covariance(self):
        """Displays the covariance results to the console

        :param dict variances_p: The dict of parameter variances to display

        :return: None

        """
        import scipy.stats as st
        
        if not hasattr(self, 'confidence') or self.confidence is None:
            confidence = 0.95 # just because
        
        number_of_stds = st.norm.ppf(1 - (1 - confidence)/2)
        margin = 15
        variances_p = self.p_variances
                        
        print(f'\n# Parameter Values (Confidence: {int(confidence*100)}%)')
        for exp in self.experiments:
            print(f'\nExperiment - {exp}:\n')
            for i, (k, p) in enumerate(self.model.experiment[exp].P.items()):
                if p.is_fixed():
                    continue
                is_global = '(local)'
                if k in self.global_params:
                    is_global = '(global)'
                
                key = f'P[{k}]'
                variance = self.rm_variances[exp][k]
                if k in self.global_params:
                    print(f'{k.rjust(margin)} = {p.value:0.4e} +/- {number_of_stds*(variance**0.5):0.4e}    {is_global}')
                
            # These need to be set equal to the local values
            if hasattr(self.model.experiment[exp], 'Pinit'):
                for i, k in enumerate(self.model.experiment[exp].Pinit.keys()):
                    self.model.experiment[exp].Pinit[k] = self.model.experiment[exp].init_conditions[k].value
                    key = f'Pinit[{k}]'
                    value = self.model.experiment[exp].Pinit[k].value
                    variance = self.rm_variances[exp][k]
                    print_name = f'{k} (init)'
                    print(f'{print_name.rjust(margin)} = {value:0.4e} +/- {number_of_stds*(variance)**0.5:0.4e}    (local)')
                        
            # These need to be set equal to the local values
            if hasattr(self.model.experiment[exp], 'time_step_change'):
                for i, k in enumerate(self.model.experiment[exp].time_step_change.keys()):
                    self.model.experiment[exp].time_step_change[k] = self.model.experiment[exp].time_step_change[k].value
                    key = f'time_step_change[{k}]'
                    value = self.model.experiment[exp].time_step_change[k].value
                    variance = self.rm_variances[exp][k]
                    print(f'{k.rjust(margin)} = {value:0.4e} +/- {number_of_stds*(variance)**0.5:0.4e}    (local)')
                          
        return None
    
    
    def covariance(self, solver, solver_factory):
        """Solves for the Hessian regardless of data source - only uses ipopt_sens

        :param str solver: The name of the solver

        :Keyword Args:
            sigma_sq (dict): variances

            optimizer (SolverFactory): Pyomo Solver factory object

            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not

        :return numpy.ndarray hessian: The hessian matrix for covariance calculations
            
        """
        components = self.reaction_models[self.experiments[0]].p_estimator.comps['unknown_absorbance']
        parameters = self.param_names_full
        models_dict = {k: v.p_model for k, v in self.reaction_models.items()}
        
        if solver == 'ipopt_sens':
            from kipet.estimator_tools.reduced_hessian_methods import covariance_sipopt
            covariance_matrix, covariance_matrix_reduced = covariance_sipopt(
                models_dict, 
                solver_factory,
                components, 
                parameters, 
                mee_obj=self.model,
            )
        
        else:
            from kipet.estimator_tools.reduced_hessian_methods import covariance_k_aug
            covariance_matrix, covariance_matrix_reduced = covariance_k_aug(
                models_dict, 
                None,
                components,
                parameters, 
                mee_obj=self.model,
            )
            
            #%%
            
        # self = lab.mee
        index = []
        for name in self.param_names_full:
            
            sep = name.split('[', 1)[1].split(']')
            exp = sep[0]
            kind = sep[1][1:].split('[')[0]
            var = sep[1][1:].split('[')[1]
            
            if var not in self.global_params:
                index.append(f'{kind}[{var}] {exp}')
            else:
                index.append(f'{kind}[{var}] {exp}')
            
        H = covariance_matrix_reduced
        if not hasattr(self.reaction_models[self.experiments[0]].p_model, 'C'):
            self.cov = pd.DataFrame(H, index=index, columns=index)  
        else:  
            from kipet.estimator_tools.reduced_hessian_methods import compute_covariance
            models_dict = {k: v.p_model for k, v in self.reaction_models.items()}
            free_params = len(self.param_names)
            all_variances = self.variances
            V_theta = compute_covariance(models_dict, H, free_params, all_variances)
            self.cov = pd.DataFrame(V_theta, index=index, columns=index)  
        
        #%%
        self.rm_variances = {}
        self.rm_cov = {}
        for name, rm in self.reaction_models.items():
            self.rm_variances[name] = {}
            
            col = []
            for i in index:
                if i.split(' ')[1] == name or i.split(' ')[0] in self.global_params_full:
                    
                    col.append(i)
                    param_name = i.split('[', 1)[1].split(']')[0]
                    self.rm_variances[name][param_name] = self.cov.loc[i, i]
            
            rm_cov = self.cov.loc[col, col]
            rm_cov.columns = [c.split('[', 1)[1].split(']')[0] for c in col]
          
            c = self.cov
            c = c.loc[col, col]
            new_cols = [c.split('[', 1)[1].split(']')[0] for c in col]
            
            c.columns = new_cols
            c.index = new_cols
            
            self.rm_cov[name] = c
        
        #%%
        
        self.p_variances = np.diag(self.cov.values)
        self._display_covariance()
        
        return None
        
    def _scale_variances(self,):
        """Option to scale the variances for MEE

        :return: None

        """
        var_scaled = dict()
        for s,t in self.variances.items():
            maxx = max(list(t.values()))
            ind_var = dict()
            for i,j in t.items():
                ind_var[i] = j/maxx
            var_scaled[s] = ind_var
        self.variances = var_scaled
        self.variance_scale = maxx
        
        return None


    def solve_consolidated_model(self, 
                                 global_params=None,
                                 **kwargs):
        """This function consolidates the individual models into a single
        optimization problem that links the parameters and spectra (if able)
        from each experiment.

        :param list global_params: This is the list of global parameters to be linked in the MEE
        :param dict kwargs: The dictionary of options passed from ReactionSet
        
        """
        solver_opts = kwargs.get('solver_opts', {'linear_solver': 'ma57'})
        tee = kwargs.get('tee', True)
        scaled_variance = kwargs.get('scaled_variance', False)
        shared_spectra = kwargs.get('shared_spectra', True)
        solver = kwargs.get('solver', 'ipopt')
        parameter_means = kwargs.get('mean_start', True)
        
        covariance = kwargs.get('covariance', None)
        
        from kipet import __version__ as version_number
        
        print('#' * 40)
        print(f'# KIPET version {version_number}')
        print(f'# Date: {list(self.reaction_models.values())[0].timestamp}')
        print(f'# Date: {list(self.reaction_models.values())[0].file.stem}')
        print(f'# ReactionModel instances: {", ".join(list(self.reaction_models.keys()))}')
        print('#' * 40)
        
        print("\n# Multiple Experiments: Starting parameter estimation \n")
       
        combined_model = ConcreteModel()
        
        self.variance_scale = 1
        if scaled_variance == True:
            self._scale_variances()

        if global_params is None:
            # This needs to be a global attr
            self.global_params = self.all_params
        else:
            self.global_params = global_params
            
        # Parameter name list
        self.global_params_full = [f'P[{p}]' for p in self.global_params]
        
        def build_individual_blocks(m, exp):
            """This function forms the rule for the construction of the individual blocks 
            for multiple experiments, referenced in run_parameter_estimation. This function 
            is not meant to be used by users directly.

            :param ConcreteModel m: The concrete model that we are adding the block to
            :param list exp: A list containing the experiments
                
            :return ConcreteModel m: Pyomo model inside the block (after modification)

            """
            list_components = self.reaction_models[exp].components.names
            with_d_vars= True
            m = copy.copy(self.reaction_models[exp].p_model)
            
            # Quick fix - I don't know what is causing this
            if hasattr(m, 'alltime_domain'):
                m.del_component('alltime_domain')
            if hasattr(m, 'huplctime_domain'):
                m.del_component('huplctime_domain')
            
            if with_d_vars and hasattr(m, 'D'):
              
                m.D_bar = Var(m.times_spectral,
                              m.meas_lambdas)
    
                def rule_D_bar(m, t, l):   
                    return m.D_bar[t, l] == sum(getattr(m, self.__var.concentration_spectra)[t, k] * getattr(m, self.__var.spectra_species)[l, k] for k in self.reaction_models[exp].p_estimator.comps['unknown_absorbance'])
    
                m.D_bar_constraint = Constraint(m.times_spectral,
                                                m.meas_lambdas,
                                                rule=rule_D_bar)
            
            m.error = Var(bounds = (0, None))
                
            def rule_objective(m):
                
                expr = 0
                spectral_term = 0
                concentration_term = 0
                measured_concentration_term = 0
                complementary_state_term = 0
                weights = [1, 1, 1, 1]
                obj_variances = self.variances
                
                if hasattr(m, self.__var.spectra_data):
                    spectral_term = absorption_objective(m, 
                                                 device_variance=obj_variances[exp]['device'],
                                                 g_option=self.reaction_models[exp]._G_data['G_contribution'],
                                                 with_d_vars=with_d_vars,
                                                 shared_spectra=shared_spectra,
                                                 species_list=list_components)
                    
                    concentration_term = conc_objective(m, variance=obj_variances[exp], source='spectra')
                
                if hasattr(m, self.__var.concentration_measured):
                    measured_concentration_term = conc_objective(m, variance=obj_variances[exp])
                
                if hasattr(m, self.__var.state):
                    complementary_state_term = comp_objective(m, variance=obj_variances[exp])
                    
                expr = weights[0]*spectral_term + \
                    weights[1]*concentration_term + \
                    weights[2]*measured_concentration_term + \
                    weights[3]*complementary_state_term
    
                return m.error == expr
    
            m.obj_const = Constraint(rule=rule_objective)
                
            return m  
        
        combined_model.experiment = Block(self.experiments, rule=build_individual_blocks)
        combined_model.map_exp_to_count = dict(enumerate(self.experiments))
        
        def param_linking_rule(m, exp, param):
            prev_exp = None
            key = next(key for key, value in combined_model.map_exp_to_count.items() if value == exp)
            if key == 0:
                return Constraint.Skip
            else:
                for key, val in combined_model.map_exp_to_count.items():
                    if val == exp:
                        prev_exp = combined_model.map_exp_to_count[key-1]
                if param in self.global_params and prev_exp != None:
                    return getattr(combined_model.experiment[exp], self.__var.model_parameter)[param] == getattr(combined_model.experiment[prev_exp], self.__var.model_parameter)[param]
                else:
                    return Constraint.Skip
        
        set_fixed_params=set()
        
        for exp in self.experiments:
            for param, param_obj in getattr(combined_model.experiment[exp], self.__var.model_parameter).items():
                if param_obj.is_fixed():
                    set_fixed_params.add(param)
            
        if len(set_fixed_params) > 0:
            print(f'# Multiple Experiments: The fixed parameters are:\n{set_fixed_params}')
        
        set_params_across_blocks = self.all_params.difference(set_fixed_params)
        combined_model.parameter_linking = Constraint(self.experiments, set_params_across_blocks, rule = param_linking_rule)
        
        def wavelength_linking_rule(m, exp, wave, comp):
            prev_exp = None
            key = next(key for key, value in combined_model.map_exp_to_count.items() if value == exp)
            if key == 0:
                return Constraint.Skip
            else:
                for key, val in combined_model.map_exp_to_count.items():
                    if val == exp:
                        prev_exp = combined_model.map_exp_to_count[key-1]
                if wave in self.all_wavelengths and prev_exp != None:
                    if comp in combined_model.experiment[prev_exp].mixture_components and comp in combined_model.experiment[exp].mixture_components:
                        return getattr(combined_model.experiment[exp], self.__var.spectra_species)[wave,comp] == getattr(combined_model.experiment[prev_exp], self.__var.spectra_species)[wave,comp]
                    else:
                        return Constraint.Skip
                else:
                    return Constraint.Skip

        if shared_spectra == True:
            combined_model.spectra_linking = Constraint(self.experiments, self.all_wavelengths, self.all_species, rule = wavelength_linking_rule)
        
        # Add in experimental weights
        combined_model.objective = Objective(sense=minimize, expr=sum(b.error for b in combined_model.experiment[:]))
        
        self.model = combined_model
        
        models = {k: v.p_model for k, v in self.reaction_models.items()}
        self.param_names = define_free_parameters(models, self.global_params_full, kind='variable')
        self.param_names_full = define_free_parameters(models, self.global_params_full, kind='full')
        
        if covariance in ['k_aug', 'ipopt_sens']:
            
            tee = kwargs.pop('tee', False)
            solver_opts = kwargs.pop('solver_opts', dict())
            optimizer = None
            
            # At the moment k_aug is not working for this
            if covariance == 'k_aug':
                covariance = 'ipopt_sens'
            
            if covariance == 'ipopt_sens':
                if not 'compute_red_hessian' in solver_opts.keys():
                    solver_opts['compute_red_hessian'] = 'yes'
                    
                # Create the optimizer
                optimizer = SolverFactory(covariance)
                for key, val in solver_opts.items():
                    optimizer.options[key] = val
            
            self.covariance(covariance, optimizer)
            
        else:
            optimizer = SolverFactory('ipopt')
            optimizer.solve(combined_model, options=solver_opts, tee=True)            
        
        solver_results = {}
        
        for i in combined_model.experiment:
            solver_results[i] = ResultsObject()
            solver_results[i].load_from_pyomo_model(combined_model.experiment[i])
            if self.rm_cov is not None:
                solver_results[i].parameter_covariance = self.rm_cov[i]
            
            #setattr(solver_results[i], 'variances', self.rm_variances[i])
            
        self.results = solver_results                     
        print('\n# Multiple Experiments: Parameter estimation complete\n')
                
        return solver_results
    
    @property
    def all_params(self):
        """Method to return all of the parameters in the ReactionModels

        :return set set_of_all_model_params: The set of all parameters

        """
        set_of_all_model_params = set()
        for name, model in self.reaction_models.items():
            set_of_all_model_params = set_of_all_model_params.union(model.parameters.names)
        return set_of_all_model_params
    
    
    @property
    def all_wavelengths(self):
        """Method to return all of the wavelengths in the ReactionModels

        :return set set_of_all_wavelengths: The set of all wavelengths

        """
        set_of_all_wavelengths = set()
        for name, model in self.reaction_models.items():
            set_of_all_wavelengths = set_of_all_wavelengths.union(list(model.p_model.meas_lambdas))
        return set_of_all_wavelengths
    
    @property
    def all_species(self):
        """Method to return all of the components in the ReactionModels

        :return set set_of_all_species: The set of all components

        """
        set_of_all_species = set()
        for name, model in self.reaction_models.items():
            set_of_all_species = set_of_all_species.union(model.components.names)
        return set_of_all_species
